import numpy as np
from skimage import io
from scipy.optimize import curve_fit
from scipy.stats import norm
import matplotlib.pyplot as plt
from matplotlib import cm
from matplotlib.ticker import LinearLocator, FormatStrFormatter
from xcorr_funcs import normxcorr2, FastPeakFind, maxk, gauss1
from scipy import exp
import time

def track_spheres(img, mask, num_particles, subpixel=True):
    def gauss1(x,a,x0,sigma):
        return a*exp(-(x-x0)**2/(2*sigma**2))
    corr = normxcorr2(mask, img, mode='same')
    cent = FastPeakFind(corr)
    peaks = corr[cent[0], cent[1]]
    ind = maxk(peaks, num_particles)
    max_coor_tmp = cent[:, ind]
    max_coor = max_coor_tmp.astype('float32')
    pk_value = peaks[ind]
    if subpixel == True:
        for num in range(0, num_particles):
            x = max_coor_tmp[0, num]
            y = max_coor_tmp[1, num]
            fitx1 = np.asarray(range(x-7, x+8))
            fity1 = np.asarray(corr[range(x-7, x+8), y])
            popt, pcov = curve_fit(gauss1, fitx1, fity1, p0=[1, x, 3])
            max_coor[0, num] = popt[1]
            fitx2 = np.asarray(range(y-7, y+8))
            fity2 = np.asarray(corr[x, range(y-7, y+8)])
            popt,pcov = curve_fit(gauss1, fitx2, fity2, p0=[1, y, 3])
            max_coor[1, num] = popt[1]
    return max_coor, pk_value

if __name__ == '__main__':
    t1 = time.monotonic()
    img = io.imread('video.tif')
    mask = io.imread('maski.tif')
    num_images = img.shape[0]
    num_particles = 3
    nTotal = 100
    fid = open('xyt.dat', 'w')
    fid.write('X\tY\tframe\n')
    for num, frame in enumerate(img):
        print('Processing image ' + str(num) + ' ...')
        max_coor, pk_value = track_spheres(frame, mask, num_particles)
        for coor in max_coor.transpose():
            fid.write('%f\t%f\t%d\n' % (coor[0], coor[1], num))
        # plt.imshow(frame)
        # plt.plot(max_coor[1], max_coor[0], 'ro')
        # plt.show('block')
        if num >= nTotal - 1:
            break
    fid.close()
    t2 = time.monotonic()
    print('Python processed ' + str(nTotal) + ' images in ' + str(t2-t1) + ' secs')



# a set of functions for the circle fitting GUI
import numpy as np
import matplotlib.pyplot as plt
import os
from nd2reader import ND2Reader
from skimage import io
import matplotlib.pyplot as plt
import pandas as pd
import matplotlib as mpl
import ctypes
from scipy.optimize import minimize

"""
Drying picture analyzing app utility functions
"""


def determine_figsize(img, ratio=0.9, dpi=100):
    """
    Determine the display size of image

    Args:
    img -- the image to display
    ratio -- the maximum fraction the image can take up, beyond which the image is rescaled
    dpi -- dot per inch

    Returns:
    shape -- determined figsize
    compressRatio -- how much image gets compressed

    Test:
    >>> img = io.imread(r'E:\Google Drive\Pictures and videos\research\DD\crack_pattern_of_bacterial_suspension_droplet.png', as_gray=True)
    >>> shape, cr = determine_figsize(img, ratio=0.9, dpi=100)
    """
    h, w = img.shape
    hcanvas, wcanvas = h, w
    compressRatio = 1
    user32 = ctypes.windll.user32
    wmax = np.floor(ratio*user32.GetSystemMetrics(0))
    hmax = np.floor(ratio*user32.GetSystemMetrics(1))
    if wcanvas > wmax:
        wcanvas = wmax
        hcanvas = h/w*wcanvas
        compressRatio = wmax/w
    if hcanvas > hmax:
        hcanvas = hmax
        wcanvas = w/h*hcanvas
        compressRatio = hmax/h

    return (wcanvas, hcanvas), compressRatio

def center_normalize_xy(x, y):
    """
    center and normalize x and y, to make position 0 mean and make length order 1.

    Args:
    x, y -- array-like, same length

    Returns:
    xr, yr -- centered and normalized x and y
    cache -- save the constant that will be used to scale back x and y

    Test:
    >>> xv = np.array([1, 1, -1, -0.8, 0.5])
    >>> yv = np.array([1, -1, -1, 0.8, -0.9])
    >>> center_normalize_xy(xv, yv)
    """

    assert(len(x)==len(y))

    xr = x - x.mean()
    yr = y - y.mean()
    S = ((xr**2).mean() + (yr**2).mean())**0.5
    xr = xr / S
    yr = yr / S
    cache = (x.mean(), y.mean(), S)

    return xr, yr, cache

def compute_gradient(params, x, y):
    """
    Compute the gradient of objective function at given point params

    Args:
    params -- the point where the gradient is evaluated

    Returns:
    grad -- gradients at the given point
    """

    assert(len(x)==len(y))
    a = params['a']
    b = params['b']

    r = ((x - a)**2 + (y-b)**2)**0.5
    u = (x - a) / r
    v = (y - b) / r
    da = 2 * (a + u.mean() * r.mean())
    db = 2 * (b + v.mean() * r.mean())

    grads = {'a': da, 'b': db}

    return grads

def update_params(params, grads, updating_rate):
    """
    Use grads to update params.

    Args:
    params -- fitting parameters
    grads -- gradient evaluated at params, by compute_gradient()
    updating_rate -- the speed of updating the parameters

    Returns:
    updated_params -- updated parameters
    """

    updated_params = {}

    for kw in params:
        updated_params[kw] = params[kw] - updating_rate * grads[kw]

    return updated_params

def scale_back(params, cache):
    """
    Scale back the parameters which were rescaled by center_normalize_xy()

    Args:
    params -- a and b, position of original points
    cache -- cached values of center_normalize_xy()

    Returns:
    rescaled_params -- rescaled parameters

    Test:
    >>> params = {'a': 0, 'b': 0.0}
    >>> cache = (1, 2, 3)
    >>> scale_back(params, cache)
    """

    xmean, ymean, S = cache
    a = params['a'] * S + xmean
    b = params['b'] * S + ymean

    rescaled_params = {'a': a, 'b': b}

    return rescaled_params

def compute_radius(params, x, y):
    """
    Calculate radius when center of circle is known

    Args:
    params -- center of circle
    x, y -- a set of points on edge of circle

    Returns:
    radius -- radius of circle
    """

    radius = (((x - params['a'])**2 + (y - params['b'])**2)**0.5).mean()

    return radius

def compute_hessian(params, x, y):
    """
    Compute the gradient of objective function at given point params

    Args:
    params -- the point where the gradient is evaluated
    x, y -- two coords of given points

    Returns:
    hess -- hessian matrix at the given point
    """

    assert(len(x)==len(y))
    a = params['a']
    b = params['b']

    r = ((x - a)**2 + (y-b)**2)**0.5
    u = (x - a) / r
    v = (y - b) / r

    h11 = 1 - u.mean() ** 2 - r.mean() * (v ** 2 / r).mean()
    h12 = - u.mean() * v.mean() + r.mean() * (u*v/r).mean()
    h21 = h12
    h22 = 1 - v.mean() ** 2 - r.mean() * (u ** 2 / r).mean()

    hess = 2 * np.array([[h11, h12],
                         [h21, h22]])
    return hess

def update_params_abdul(params, grads, hessian, updating_rate):
    """
    Use grads to update params.

    Args:
    params -- fitting parameters
    grads -- gradient evaluated at params, by compute_gradient()
    hessian -- hessian matrix, computed by compute_hessian()
    updating_rate -- the speed of updating the parameters

    Returns:
    updated_params -- updated parameters
    """

    hess_lambda = hessian + updating_rate * np.identity(2)
    grads_mat = np.array([grads["a"], grads["b"]])
    hmat = - np.matmul(np.linalg.inv(hess_lambda), grads_mat)
    h = {"a": hmat[0], "b": hmat[1]}
    updated_params = {}

    for kw in params:
        updated_params[kw] = params[kw] + h[kw]

    return updated_params

def fit_circle(x, y, updating_rate=0.1, method="naive"):
    """
    Fit a set of points x and y with a circle (a, b, r), using gradient descent method

    Args:
    x, y -- coordinates of a set of points
    updating_rate -- rate of gradient descent
    method -- "naive": use lambda*grads to update params
              "abdul": use hessian matrix
              "linear": use linear method (Coope 1993)

    Returns:
    output_params -- parameters of fitted circle

    Edit:
    07132022 -- Implement "abdul" and "linear" methods, based on Abdul 2014 and Coope 1993
    """
    def fun(y, *args):
        return ((np.matmul(args[0], y) - args[1]) ** 2).sum()
    if method == "linear":
        B = np.stack((x, y, np.ones(x.shape)), axis=1)
        d = x ** 2 + y ** 2
        x0 = np.array([x.mean(), y.mean(), 0])
        res = minimize(fun, x0, args=(B, d))
        y = res["x"]
        a, b = y[0]/2, y[1]/2 # recover results
        r = (y[2] + a**2 + b**2) ** 0.5
        return {"a": a, "b": b, "r": r}, res["nit"]

    xr, yr, cache = center_normalize_xy(x, y)
    params = {'a': xr.mean(), 'b': yr.mean()}
    grad_norm = 1
    n_iter = 0
    while grad_norm > 3e-8:
        grads = compute_gradient(params, xr, yr)
        if method == "naive":
            params = update_params(params, grads, updating_rate)
        elif method == "abdul":
            hess = compute_hessian(params, xr, yr)
            params = update_params_abdul(params, grads, hess, updating_rate)
        grad_norm = (grads['a']**2 + grads['b']**2)**0.5
        n_iter += 1
    output_params = scale_back(params, cache)
    output_params['r'] = compute_radius(output_params, x, y)

    return output_params, n_iter

if __name__ == '__main__':
    img = io.imread(r'E:\Google Drive\Pictures and videos\research\DD\crack_pattern_of_bacterial_suspension_droplet.png', as_gray=True)
    shape, cr = determine_figsize(img, ratio=0.9, dpi=100)
    print(shape)



"""
GENERAL
=======
Double emulsion project code library.
"""
import numpy as np
import pandas as pd
import json
import matplotlib.pyplot as plt
from matplotlib.patches import Ellipse
from matplotlib.backends.backend_agg import FigureCanvasAgg as FigureCanvas
from matplotlib.figure import Figure
from skimage import io
from myimagelib.myImageLib import readdata
from myimagelib.pivLib import read_piv, PIV, PIV_masked
import corrTrack
import os
import trackpy as tp

# %% codecell

class droplet_image:
    """Container of functions related to confocal droplet images"""
    def __init__(self, image_sequence, fps, mpp):
        """image_sequence: dataframe of image dir info, readdata return value
        mask: image mask, the same as PIV mask
        xy0: 2-tuple of droplet initial coordinates, read from positions.csv file
        mask_shape: shape of the circular mask, 2-tuple specifying the rect bounding box of the mask, typically a square
                    read from positions.csv file"""
        self.sequence = image_sequence
        self.fps = fps
        self.mpp = mpp
    def __repr__(self):
        I0 = io.imread(self.sequence.iloc[0, 1])
        return "length: {0:d}, image size: {1}, fps: {2:.1f}, mpp: {3:.2f}".format(
                len(self.sequence), str(I0.shape), self.fps, self.mpp)
    def __len__(self):
        return len(self.sequence)
    def process_first_image(self, mask):
        """Compare the provided mask (template) with the first image.
        If the maximum correlation is at the center of the image, the input mask should be correct.
        This function also records the measured center coordinates (xc, yc), for computing the offset."""
        img = io.imread(self.sequence.Dir[0])
        h, w = img.shape
        xy, pkv = corrTrack.track_spheres(img, mask, 1, subpixel=True)
        xyc = np.flip(xy.squeeze())
        # The center coords should be very close to the image center (w/2, h/2)
        diff = np.square(xyc-np.array([w/2, h/2])).sum()
        if diff > 10: # this threshold is arbitrary
            print("The detected center is {:.1f} pixels from the image center, check if the input mask is correct.".format(diff**0.5))
        return xyc
    def droplet_traj(self, mask, xy0):
        xyc = self.process_first_image(mask)
        xym_list = []
        for num, i in self.sequence.iterrows():
            print("Tracking droplet: {}".format(self.get_image_name(num)), end="\r")
            img = io.imread(i.Dir)
            xy, pkv = corrTrack.track_spheres(img, mask, 1)
            xym_list.append(np.flip(xy.squeeze()))
        xym = np.stack(xym_list, axis=0)
        traj = self.sequence.copy()
        traj = traj.assign(x=xy0[0] + xym[:, 0] - xyc[0], y=xy0[1] + xym[:, 1] - xyc[1])
        self.traj = traj
        return traj
    def get_cropped_image(self, index, traj, mask_shape):
        """Retrieve cropped image by index.
        Mar 10, 2022 -- boundary protection."""
        x = int(traj.x[index])
        y = int(traj.y[index])
        h, w = mask_shape
        x0 = x - w // 2
        x1 = x0 + w
        y0 = y - h // 2
        y1 = y0 + h
        img = self.get_image(index)
        ih, iw = img.shape
        cropped = img[max(y0, 0): min(y1, ih), max(x0, 0): min(x1, iw)]
        return cropped
    def get_image(self, index):
        """Retrieve image by index"""
        imgDir = self.sequence.Dir[index]
        img = io.imread(imgDir)
        return img
    def check_traj(self, traj, mask_shape, n=10):
        """Check the droplet finding.
        traj: DataFrame containing columns 'x' and 'y'
        mask_shape: 2-tuple, (width, height)
        n: total number of frames to inspect, default to 10"""
        interval = len(traj) // n
        count = 0
        for num, i in traj[::interval].iterrows():
            fig, ax = plt.subplots()
            img = io.imread(i.Dir)
            elli = Ellipse((i.x, i.y), *mask_shape, facecolor=(0,0,0,0), lw=1, edgecolor="red")
            ax.imshow(img, cmap="gray")
            ax.add_patch(elli)
            count += 1
            if count >= n:
                break
    def get_image_name(self, index):
        return self.sequence.Name[index]
    def fixed_mask_piv(self, winsize, overlap, mask_dir):
        """Edit:
        08092022 -- Return PIV results instead of saving to file.
                    Remove arg save_folder .
                    Remove arg dt, add required arg fps in __init__ instead
        """
        dt = 1 / self.fps
        mask = io.imread(mask_dir)
        data = {}
        for i0, i1 in zip(self.sequence.index[::2], self.sequence.index[1::2]):
            I0 = self.get_image(i0)
            I1 = self.get_image(i1)
            x, y, u, v = PIV_masked(I0, I1, winsize, overlap, dt, mask)
            # generate dataframe and save to file
            tmp = pd.DataFrame({"x": x.flatten(), "y": y.flatten(), "u": u.flatten(), "v": v.flatten()})
            data["{0}-{1}".format(self.get_image_name(i0), self.get_image_name(i1))] = tmp
            # data.to_csv(os.path.join(save_folder, "{0}-{1}.csv".format(self.get_image_name(i0), self.get_image_name(i1))), index=False)
        params = {"winsize": winsize,
                  "overlap": overlap,
                  "dt": dt,
                  "mask_dir": mask_dir}
        # with open(os.path.join(save_folder, "piv_params.json"), "w") as f:
        #     json.dump(params, f)
        return data, params
    def piv_overlay_fixed(self, piv_folder, out_folder, sparcity=1):
        """Draw PIV overlay for fixed mask PIV data (unify old code in class)"""
        def determine_arrow_scale(u, v, sparcity):
            row, col = u.shape
            return max(np.nanmax(u), np.nanmax(v)) * col / sparcity / 1.5
        if os.path.exists(out_folder) == False:
            os.makedirs(out_folder)
        l = readdata(piv_folder, "csv")
        # determine scale using the first frame
        x, y, u, v = read_piv(l.Dir[0])
        scale = determine_arrow_scale(u, v, sparcity)

        for num, i in l.iterrows():
            name = i.Name.split("-")[0]
            index = self.sequence.loc[self.sequence.Name==name].index[0]
            img = self.get_image(index)
            x, y, u, v = read_piv(i.Dir)
            # sparcify
            row, col = x.shape
            xs = x[0:row:sparcity, 0:col:sparcity]
            ys = y[0:row:sparcity, 0:col:sparcity]
            us = u[0:row:sparcity, 0:col:sparcity]
            vs = v[0:row:sparcity, 0:col:sparcity]
            # plot quiver
            dpi = 300
            figscale = 1
            w, h = img.shape[1] / dpi, img.shape[0] / dpi
            fig = Figure(figsize=(w*figscale, h*figscale)) # on some server `plt` is not supported
            canvas = FigureCanvas(fig) # necessary?
            ax = fig.add_axes([0, 0, 1, 1])
            ax.imshow(img, cmap='gray')
            ax.quiver(xs, ys, us, vs, color='yellow', width=0.003, \
                        scale=scale, scale_units='width') # it's better to set a fixed scale, see *Analysis of Collective Motions in Droplets* Section IV.A.2 for more info.
            ax.axis('off')
            # save figure
            fig.savefig(os.path.join(out_folder, name + '.jpg'), dpi=dpi)
    def moving_mask_piv(self, save_folder, winsize, overlap, dt, mask_dir, xy0, mask_shape):
        """Perform moving mask PIV and save PIV results and parameters in save_folder"""
        # create save_folder
        if os.path.exists(save_folder) == False:
            os.makedirs(save_folder)
        # save params in .json file, so that only PIV data are saved in .csv files
        params = {"winsize": winsize,
                  "overlap": overlap,
                  "dt": dt,
                  "mask_dir": mask_dir,
                  "droplet_initial_position (xy0)": (int(xy0[0]), int(xy0[1])),
                  "mask_shape": (int(mask_shape[0]), int(mask_shape[1]))}
        # save traj data in .json file, so that only PIV data are saved in .csv files
        with open(os.path.join(save_folder, "piv_params.json"), "w") as f:
            json.dump(params, f)
        # track droplet
        mask = io.imread(mask_dir)
        traj = self.droplet_traj(mask, xy0)
        traj.to_json(os.path.join(save_folder, "droplet_traj.json"))
        # PIV
        print("")
        n = 0
        for i0, i1 in zip(self.sequence.index[::2], self.sequence.index[1::2]):
            name0 = self.get_image_name(i0)
            name1 = self.get_image_name(i1)
            print("PIV: {0}-{1}".format(name0, name1), end="\r")
            I0 = self.get_cropped_image(i0, traj, mask_shape)
            I1 = self.get_cropped_image(i1, traj, mask_shape)
            x, y, u, v = PIV(I0, I1, winsize, overlap, dt)
            # apply (circular) mask to u and v: implement in the piv_data class

            # generate dataframe and save to file
            data = pd.DataFrame({"x": x.flatten(), "y": y.flatten(), "u": u.flatten(), "v": v.flatten()})
            data.to_csv(os.path.join(save_folder, "{0}-{1}.csv".format(self.get_image_name(i0), self.get_image_name(i1))), index=False)
            n += 1
            if n != 0 and n % 100 == 0:
                print("PIV: {0}-{1}".format(name0, name1))
    def piv_overlay_moving(self, piv_folder, out_folder, traj, piv_params, sparcity=1, crop=True):
        """Draw PIV overlay for moving mask piv data (only on cropped images)"""
        def determine_arrow_scale(u, v, sparcity):
            row, col = u.shape
            return max(np.nanmax(u), np.nanmax(v)) * col / sparcity / 1.5
        if os.path.exists(out_folder) == False:
            os.makedirs(out_folder)
        mask_shape = piv_params["mask_shape"]
        l = readdata(piv_folder, "csv")
        # determine scale using the first frame
        x, y, u, v = read_piv(l.Dir[0])
        scale = determine_arrow_scale(u, v, sparcity)

        for num, i in l.iterrows():
            name = i.Name.split("-")[0]
            index = self.sequence.loc[self.sequence.Name==name].index[0]
            img = self.get_cropped_image(index, traj, mask_shape)
            x, y, u, v = read_piv(i.Dir)
            # sparcify
            row, col = x.shape
            xs = x[0:row:sparcity, 0:col:sparcity]
            ys = y[0:row:sparcity, 0:col:sparcity]
            us = u[0:row:sparcity, 0:col:sparcity]
            vs = v[0:row:sparcity, 0:col:sparcity]
            # plot quiver
            dpi = 300
            figscale = 1
            w, h = img.shape[1] / dpi, img.shape[0] / dpi
            fig = Figure(figsize=(w*figscale, h*figscale)) # on some server `plt` is not supported
            canvas = FigureCanvas(fig) # necessary?
            ax = fig.add_axes([0, 0, 1, 1])
            ax.imshow(img, cmap='gray')
            ax.quiver(xs, ys, us, vs, color='yellow', width=0.003, \
                        scale=scale, scale_units='width') # it's better to set a fixed scale, see *Analysis of Collective Motions in Droplets* Section IV.A.2 for more info.
            ax.axis('off')
            # save figure
            fig.savefig(os.path.join(out_folder, name + '.jpg'), dpi=dpi)



# %% codecell
class fixed_mask_PIV:
    def PIV_masked_1(self, I0, I1, winsize, overlap, dt, mask):
        """Test different masking procedures.
        1. Mask raw images by setting background 0, then apply PIV.
        2. Mask raw images by setting background nan, then apply PIV.
        3. Apply PIV directly on raw images, then apply mask on velocity.
        """
        assert(mask.shape==I0.shape)
        mask = mask >= mask.mean() # convert mask to boolean array
        I0 = I0 * mask
        I1 = I1 * mask
        x, y, u, v = PIV(I0, I1, winsize, overlap, dt)
        mask_w = divide_windows(mask, windowsize=[winsize, winsize], step=winsize-overlap)[2] >= 1
        assert(mask_w.shape==x.shape)
        u[~mask_w] = np.nan
        v[~mask_w] = np.nan
        return x, y, u, v
    def PIV_masked_2(self, I0, I1, winsize, overlap, dt, mask):
        """Test different masking procedures.
        1. Mask raw images by setting background 0, then apply PIV.
        2. Mask raw images by setting background nan, then apply PIV.
        3. Apply PIV directly on raw images, then apply mask on velocity.
        """
        assert(mask.shape==I0.shape)
        mask = mask >= mask.mean() # convert mask to boolean array
        x, y, u, v = PIV(I0, I1, winsize, overlap, dt)
        mask_w = divide_windows(mask, windowsize=[winsize, winsize], step=winsize-overlap)[2] >= 1
        assert(mask_w.shape==x.shape)
        u[~mask_w] = np.nan
        v[~mask_w] = np.nan
        return x, y, u, v
    def PIV_masked_3(self, I0, I1, winsize, overlap, dt, mask):
        """Test different masking procedures.
        1. Mask raw images by setting background 0, then apply PIV.
        2. Mask raw images by setting background nan, then apply PIV.
        3. Apply PIV directly on raw images, then apply mask on velocity.
        """
        assert(mask.shape==I0.shape)
        mask = mask >= mask.mean() # convert mask to boolean array
        I0[~mask] = np.nan
        I1[~mask] = np.nan
        x, y, u, v = PIV(I0, I1, winsize, overlap, dt)
        mask_w = divide_windows(mask, windowsize=[winsize, winsize], step=winsize-overlap)[2] >= 1
        assert(mask_w.shape==x.shape)
        u[~mask_w] = np.nan
        v[~mask_w] = np.nan
        return x, y, u, v
    def test(self):
        I0 = io.imread(os.path.join("img", "I10.tif"))
        I1 = io.imread(os.path.join("img", "I11.tif"))
        mask = io.imread(os.path.join("img", "mask1.tif"))
        winsize = 40
        overlap = 20
        dt = 0.02
        x, y, u, v = [], [], [], []
        for func in [PIV_masked_1, PIV_masked_2]:
            x1, y1, u1, v1 = func(I0, I1, winsize, overlap, dt, mask)
            x.append(x1); y.append(y1); u.append(u1); v.append(v1)
        fig, ax = plt.subplots(nrows=1, ncols=2, dpi=200)
        for i in [0, 1]:
            ax[i].imshow(I0, cmap='gray')
            ax[i].quiver(x[i], y[i], u[i], v[i], color='yellow')
            ax[i].axis('off')
        print("The two masking procedures don't produce very different results, according to visual inspection.")
        print("I also plot the velocity distribution function below.")
        fig, ax = plt.subplots(nrows=1, ncols=2, dpi=200)
        for i in [0, 1]:
            hist, bin_edges = np.histogram(u[i][~np.isnan(u[i])], density=True)
            ax[0].plot(bin_edges[:-1], hist)
            hist, bin_edges = np.histogram(v[i][~np.isnan(v[i])], density=True)
            ax[1].plot(bin_edges[:-1], hist)
        ax[0].set_xlabel("u")
        ax[1].set_xlabel("v")
        ax[0].set_ylabel("PDF")
        print("The two methods show statistically very similar results")
        print("Test the speed of the two methods.")
        t = []
        t.append(time.monotonic())
        for func in [PIV_masked_1, PIV_masked_2]:
            x1, y1, u1, v1 = func(I0, I1, winsize, overlap, dt, mask)
            t.append(time.monotonic())
        t1 = t[1] - t[0]
        t2 = t[2] - t[1]
        plt.bar([1, 2], [t1, t2])
        plt.xticks([1, 2])
        plt.ylabel("time (s)")
        print("The second method takes longer time. So the first method is better.")
        print("Since it gives similar results while using less time.")
        I20 = io.imread(os.path.join("img", "I20.tif"))
        I21 = io.imread(os.path.join("img", "I21.tif"))
        mask2 = io.imread(os.path.join("img", "mask2.tif"))
        x, y, u, v = PIV_masked(I20, I21, 20, 10, 0.02, mask2)
        fig, ax = plt.subplots()
        ax.imshow(I20, cmap='gray')
        ax.quiver(x, y, u, v, color='yellow')
        ax.axis('off')

class de_data():
    """Double emulsion data plotting tool."""
    def __init__(self, data):
        self.data = data
    def __repr__(self):
        return(str(self.data))
    def show(self):
        print(self.data)
    def parameter_space(self, highlight_Chile_data=True):
        """D vs. d, with color coded OD"""
        # log1 = self.data.dropna(subset=["Rinfy", "t2"])
        log1 = self.data
        binsize = 10 # OD bin size
        plt.figure(figsize=(3.5,3),dpi=300)
        bin_starts = range(0, int(log1.OD.max()), binsize)
        cmap = plt.cm.get_cmap("tab10")
        for num, bs in enumerate(bin_starts):
            log2 = log1.loc[(log1.OD>bs)&(log1.OD<=bs+binsize)]
            if len(log2) == 0:
                continue
            if highlight_Chile_data == True:
                log3 = log2.loc[log2.Comment!="Chile"]
                log4 = log2.loc[log2.Comment=="Chile"]
                plt.scatter(log3.D, log3.d, color=cmap(num), label="{0:d}-{1:d}".format(bs,bs+binsize))
                plt.scatter(log4.D, log4.d, edgecolors=cmap(num), marker="^", fc=(0,0,0,0))
            else:
                plt.scatter(log2.D, log2.d, color=cmap(num), label="{0:d}-{1:d}".format(bs,bs+binsize))
        plt.xlabel("$D$ (um)")
        plt.ylabel("$d$ (um)")
        plt.xlim([0, 1.05*log1.D.max()])
        plt.ylim([0, 1.05*log1.d.max()])
        plt.plot([0, 1.05*log1.d.max()], [0, 1.05*log1.d.max()], ls=":", color="black")
        plt.legend(ncol=2, fontsize=5, loc="upper left")
    def generate_msd_repo(self, component="y", data_dir=r"..\Data\traj"):
        """Generate .jpg images for MSD repo. Takes ~2 min and save images, be careful!
        coord: the displacement component to compute, can be 'y' or 'z'.
        Edit:
        Mar 03, 2022 -- i) add z component, ii) put all DE# in image file name
        Mar 18, 2022 -- Use MPP info in log (Cristian's data are in um and is not consistent with this calculation.)"""
        mapper = {"y": "<y^2>", "z": "<x^2>"}
        log1 = self.data.dropna(subset=["OD"])
        if component == "z":
            log1 = log1.loc[log1.Plane=="XZ"]
        elif component == "y":
            pass
        else:
            raise ValueError("Invalid component, should be y or z")
        viridis = plt.cm.get_cmap('Set1', 5)
        count = 0
        plt.figure(dpi=150)
        name_list = []
        for num, i in log1.iterrows():
            traj_dir = os.path.join(data_dir, "{:02d}.csv".format(int(i["DE#"])))
            if os.path.exists(traj_dir):
                traj = pd.read_csv(traj_dir)
            else:
                print("Missing traj {:d}".format(i["DE#"]))
                continue
            msd = tp.msd(traj, mpp=i.MPP, fps=i.FPS, max_lagtime=traj.frame.max()//5).dropna()
            plt.plot(msd.lagt, msd[mapper[component]], label=i["DE#"], color=viridis(count/4))
            count += 1
            name_list.append("{:d}".format(int(i["DE#"])))
            if count > 4:
                plt.legend(fontsize=20, frameon=False)
                plt.xlabel("$\Delta t$ (s)")
                if component == "y":
                    plt.ylabel(r"$\left< \Delta y^2 \right>$ ($\mu$m$^2$)")
                elif component == "z":
                    plt.ylabel(r"$\left< \Delta z^2 \right>$ ($\mu$m$^2$)")
                plt.grid(which="both", ls=":")
                plt.loglog()
                plt.savefig("{}.jpg".format("-".join(name_list)))
                plt.figure(dpi=150)
                count = 0
                name_list = []
        plt.legend(fontsize=20, frameon=False)
        plt.xlabel("$\Delta t$ (s)")
        if component == "y":
            plt.ylabel(r"$\left< \Delta y^2 \right>$ ($\mu$m$^2$)")
        elif component == "z":
            plt.ylabel(r"$\left< \Delta z^2 \right>$ ($\mu$m$^2$)")
        plt.grid(which="both", ls=":")
        plt.loglog()
        plt.tight_layout()
        plt.savefig("{:d}.jpg".format(num))
    def scatter_0(self, mode="log", highlight_Chile_data=True):
        """Plot tau^* vs. (D-d)/d^2"""
        log1 = self.data.dropna(subset=["Rinfy", "t2"])
        binsize = 20 # OD bin size
        plt.figure(figsize=(3.5,3), dpi=100)
        bin_starts = range(0, int(log1.OD.max()), binsize)
        cmap = plt.cm.get_cmap("tab10")
        for num, bs in enumerate(bin_starts):
            log2 = log1.loc[(log1.OD>bs)&(log1.OD<=bs+binsize)]
            if highlight_Chile_data == True:
                log3 = log2.loc[log2.Comment!="Chile"]
                log4 = log2.loc[log2.Comment=="Chile"]
                plt.scatter(log3["(D-d)/d^2"], log3.t2_fit, color=cmap(num), label="{0:d}-{1:d}".format(bs,bs+binsize))
                plt.scatter(log4["(D-d)/d^2"], log4.t2_fit, edgecolors=cmap(num), marker="^", fc=(0,0,0,0))
            else:
                plt.scatter(log2["(D-d)/d^2"], log2.t2_fit, color=cmap(num), label="{0:d}-{1:d}".format(bs,bs+binsize))
        plt.xlabel("$(D-d)/d^2$")
        plt.ylabel("$\\tau^*$ (s)")
        plt.legend(ncol=2, fontsize=6)
        plt.grid(which="both", ls=":")
        if mode == "log":
            plt.loglog()
    def look_for_missing_traj(self, traj_folder, fmt="{:02d}.csv"):
        """Check the existence of trajectory data file in given folder, according to the log"""
        log1 = self.data.dropna(subset=["OD"])
        n_missing = 0
        for num, i in log1.iterrows():
            traj_dir = os.path.join(traj_folder, fmt.format(int(i["DE#"])))
            if os.path.exists(traj_dir):
                traj = pd.read_csv(traj_dir)
            else:
                print("Missing traj {:d}".format(i["DE#"]))
                n_missing += 1
        print("{:d} trajectories are missing".format(n_missing))
    def plot_MSD_model_Cristian(self):
        """plot the MSD model, get a feeling of parameters, save for future use, not finished"""
        gamma = 2
        nu = 1
        t = np.logspace(-2, 2)
        y2 = (1 - np.exp(-2*gamma*t)) / (2*gamma) - (np.exp(-(gamma+nu)*t)-np.exp(-2*gamma*t))
        gamma = 0.5
        y3 = (1 - np.exp(-2*gamma*t)) / (2*gamma) - (np.exp(-(gamma+nu)*t)-np.exp(-2*gamma*t))
        plt.figure(figsize=(3.5, 3), dpi=100)
        plt.plot(t, y2, label="$\\tau=1, \\tau^*=0.5$")
        plt.plot(t, y3, label="$\\tau=1, \\tau^*=2$")
        plt.legend()
        plt.loglog()
        plt.grid(which="both", ls=":")
        plt.xlabel("lag time")
        plt.ylabel("$\left<\Delta y^2\\right>$")
    def plot_0(self, nbins=5, overlap=0, mode="log"):
        """tau vs. (D-d)/d^2, with average"""
        log = self.data
        xm = 0
        ym = 0
        cmap = plt.cm.get_cmap("tab10")
        plt.figure(figsize=(3.5,3), dpi=100)
        for num, OD_min in enumerate(range(0, 160, 20)):
            OD_max = OD_min + 20
            log1 = log.loc[(log.OD>=OD_min)&(log.OD<=OD_max)].dropna(subset=["Rinfy"])
            r = (log1.D - log1.d) / log1.d ** 2
            # visualize the bins
            # plt.figure(dpi=100)
            # plt.scatter(r, log1["DE#"])
            # for num, i in log1.iterrows():
            #     plt.annotate(i["DE#"], ((i.D - i.d) / i.d ** 2, i["DE#"]), xycoords="data")
            # plt.xlabel("$(D-d)/d^2$")
            # plt.ylabel("DE index")
            if mode == "log":
                bins = np.logspace(np.log(r.min()), np.log(r.max()), nbins+1)
            else:
                bins = np.linspace(r.min(), r.max(), nbins+1)
            bin_start = bins[:-1]
            bin_size = bins[1:] - bins[:-1]
            bin_end = bin_start + bin_size * (1 + overlap)
            # count = 20
            # for start, end in zip(bin_start, bin_end):
            #     plt.plot([start, end], [count, count])
            #     count += 2
            # plot Rinf as a function of B
            count = 0
            for start, end in zip(bin_start, bin_end):
                log2 = log1.loc[(r>=start)&(r<=end)]
                r2 = (log2.D - log2.d) / log2.d ** 2
                x = r2.mean()
                y =(log2.t2).mean()
                xe = r2.std()
                ye = (log2.t2).std()
                if count == 0:
                    plt.errorbar(x, y, xerr=xe, yerr=ye, marker="o", color=cmap(num), label="{0:d}-{1:d}".format(OD_min, OD_max))
                    count += 1
                else:
                    plt.errorbar(x, y, xerr=xe, yerr=ye, marker="o", color=cmap(num))
                if np.isnan(xe):
                    xe = 0
                if np.isnan(ye):
                    ye = 0
                if x + xe > xm:
                    xm = x + xe
                if y + ye > ym:
                    ym = y + ye
        plt.xlabel("$(D-d)/d^2$")
        plt.ylabel("$\\tau^*$ (s)")
        plt.legend(ncol=2, fontsize=6, loc="lower right")
        if mode == "log":
            plt.loglog()
            plt.grid(which="both", ls=":")
        else:
            plt.xlim([0, xm*1.1])
            plt.ylim([0, ym*1.1])
    def scatter_1(self, mode="log", highlight_Chile_data=True):
        """R_inf vs. (D-d)/d^2"""
        log1 = self.data.dropna(subset=["Rinfy", "t2"])
        binsize = 20 # OD bin size
        plt.figure(figsize=(3.5,3), dpi=100)
        bin_starts = range(0, int(log1.OD.max()), binsize)
        cmap = plt.cm.get_cmap("tab10")
        for num, bs in enumerate(bin_starts):
            log2 = log1.loc[(log1.OD>bs)&(log1.OD<=bs+binsize)]
            if highlight_Chile_data == True:
                log3 = log2.loc[log2.Comment!="Chile"]
                log4 = log2.loc[log2.Comment=="Chile"]
                plt.scatter(log3["(D-d)/d^2"], log3.Rinfy**0.5, color=cmap(num), label="{0:d}-{1:d}".format(bs,bs+binsize))
                plt.scatter(log4["(D-d)/d^2"], log4.Rinfy**0.5, edgecolors=cmap(num), marker="^", fc=(0,0,0,0))
            else:
                plt.scatter(log2["(D-d)/d^2"], log2.Rinfy**0.5, color=cmap(num), label="{0:d}-{1:d}".format(bs,bs+binsize))
        plt.xlabel("$(D-d)/d^2$")
        plt.ylabel("$R_\infty$ (um)")
        plt.legend(ncol=2, fontsize=6)
        plt.grid(which="both", ls=":")
        if mode == "log":
            plt.loglog()
    def plot_1(self, nbins=5, overlap=0, mode="log"):
        """R_inf vs. (D-d)/d^2, with average"""
        log = self.data
        xm = 0
        ym = 0
        cmap = plt.cm.get_cmap("tab10")
        plt.figure(figsize=(3.5,3), dpi=100)
        for num, OD_min in enumerate(range(0, 160, 20)):
            OD_max = OD_min + 20
            log1 = log.loc[(log.OD>=OD_min)&(log.OD<=OD_max)].dropna(subset=["Rinfy"])
            r = (log1.D - log1.d) / log1.d ** 2
            # visualize the bins
            # plt.figure(dpi=100)
            # plt.scatter(r, log1["DE#"])
            # for num, i in log1.iterrows():
            #     plt.annotate(i["DE#"], ((i.D - i.d) / i.d ** 2, i["DE#"]), xycoords="data")
            # plt.xlabel("$(D-d)/d^2$")
            # plt.ylabel("DE index")
            if mode == "log":
                bins = np.logspace(np.log(r.min()), np.log(r.max()), nbins+1)
            else:
                bins = np.linspace(r.min(), r.max(), nbins+1)
            bin_start = bins[:-1]
            bin_size = bins[1:] - bins[:-1]
            bin_end = bin_start + bin_size * (1 + overlap)
            # count = 20
            # for start, end in zip(bin_start, bin_end):
            #     plt.plot([start, end], [count, count])
            #     count += 2
            # plot Rinf as a function of B
            count = 0
            for start, end in zip(bin_start, bin_end):
                log2 = log1.loc[(r>=start)&(r<=end)]
                r2 = (log2.D - log2.d) / log2.d ** 2
                x = r2.mean()
                y =(log2.Rinfy**0.5).mean()
                xe = r2.std()
                ye = (log2.Rinfy**0.5).std()
                if count == 0:
                    plt.errorbar(x, y, xerr=xe, yerr=ye, marker="o", color=cmap(num), label="{0:d}-{1:d}".format(OD_min, OD_max))
                    count += 1
                else:
                    plt.errorbar(x, y, xerr=xe, yerr=ye, marker="o", color=cmap(num))
                if np.isnan(xe):
                    xe = 0
                if np.isnan(ye):
                    ye = 0
                if x + xe > xm:
                    xm = x + xe
                if y + ye > ym:
                    ym = y + ye
        plt.xlabel("$(D-d)/d^2$")
        plt.ylabel("$R_\infty$ (um)")
        plt.legend(ncol=2, fontsize=6, loc="lower right")
        if mode == "log":
            plt.loglog()
            plt.grid(which="both", ls=":")
        else:
            plt.xlim([0, xm*1.1])
            plt.ylim([0, ym*1.1])
    def Rinf2_tau(self):
        """Plot $R_\infty^2$ vs. $\tau^*$"""
        log = self.data
        log1 = log.dropna(subset=["Rinfy", "t2"])
        binsize = 20 # OD bin size
        plt.figure(figsize=(3.5,3), dpi=100)
        bin_starts = range(0, int(log1.OD.max()), binsize)
        cmap = plt.cm.get_cmap("tab10")
        for num, bs in enumerate(bin_starts):
            log2 = log1.loc[(log1.OD>bs)&(log1.OD<=bs+binsize)]
            log3 = log2.loc[log2.Comment!="Chile"]
            log4 = log2.loc[log2.Comment=="Chile"]
            plt.scatter(log3.t2, log3.Rinfy, color=cmap(num), label="{0:d}-{1:d}".format(bs,bs+binsize))
            plt.scatter(log4.t2, log4.Rinfy, edgecolors=cmap(num), marker="^", fc=(0,0,0,0))
        plt.xlabel("$\\tau^*$ (s)")
        plt.ylabel("$R_\infty^2 $ (um$^2$)")
        plt.legend(ncol=2, fontsize=6, loc="lower right")
        plt.grid(which="both", ls=":")
        plt.xlim([1, 30])
        plt.loglog()
    def Rinf2_over_tau(self, x="(D-d)/d^2", ax=None, xlabel=None):
        """Plot $R_\infty^2 / \tau^*$ vs. $(D-d)/d^2$
        Edit:
        05032022 -- add custom x-axis"""
        log = self.data
        log1 = log.dropna(subset=["Rinfy", "t2"])
        binsize = 20 # OD bin size
        if ax == None:
            fig, ax = plt.subplots(figsize=(3.5,3), dpi=100)
        bin_starts = range(0, int(log1.OD.max()), binsize)
        cmap = plt.cm.get_cmap("tab10")
        for num, bs in enumerate(bin_starts):
            log2 = log1.loc[(log1.OD>bs)&(log1.OD<=bs+binsize)]
            log3 = log2.loc[log2.Comment!="Chile"]
            log4 = log2.loc[log2.Comment=="Chile"]
            ax.scatter(log3[x], log3.Rinfy/(log3.t2), color=cmap(num), label="{0:d}-{1:d}".format(bs,bs+binsize))
            ax.scatter(log4[x], log4.Rinfy/(log4.t2), edgecolors=cmap(num), marker="^", fc=(0,0,0,0))
        if xlabel == None:
            xlabel = x
        ax.set_xlabel("${}$".format(xlabel))
        ax.set_ylabel("$R_\infty^2 / \\tau^*$")
        ax.legend(ncol=2, fontsize=6, loc="lower right")
        ax.grid(which="both", ls=":")
        ax.loglog()
    def rescale_Rinf_OD(self):
        """Plot Rinf/OD vs. (D-d)/d^2"""
        log = self.data
        log1 = log.dropna(subset=["Rinfy", "t2"])
        binsize = 20 # OD bin size
        plt.figure(figsize=(3.5,3), dpi=100)
        bin_starts = range(0, int(log1.OD.max()), binsize)
        cmap = plt.cm.get_cmap("tab10")
        for num, bs in enumerate(bin_starts):
            log2 = log1.loc[(log1.OD>bs)&(log1.OD<=bs+binsize)]
            log3 = log2.loc[log2.Comment!="Chile"]
            log4 = log2.loc[log2.Comment=="Chile"]
            plt.scatter(log3["(D-d)/d^2"], log3.Rinfy**0.5/(log3.OD), color=cmap(num), label="{0:d}-{1:d}".format(bs,bs+binsize))
            plt.scatter(log4["(D-d)/d^2"], log4.Rinfy**0.5/(log4.OD), edgecolors=cmap(num), marker="^", fc=(0,0,0,0))
        plt.xlabel("$(D-d)/d^2$")
        plt.ylabel("$R_\infty / OD$")
        plt.legend(ncol=2, fontsize=6)
        plt.grid(which="both", ls=":")
        plt.loglog()
    def rescale_Rinf_freespace(self):
        # rescale Rinf with (D-d)
        log = self.data
        log1 = log.dropna(subset=["Rinfy", "t2"])
        binsize = 20 # OD bin size
        plt.figure(figsize=(3.5, 3), dpi=100)
        bin_starts = range(0, int(log1.OD.max()), binsize)
        cmap = plt.cm.get_cmap("tab10")
        for num, bs in enumerate(bin_starts):
            log2 = log1.loc[(log1.OD>bs)&(log1.OD<=bs+binsize)]
            log3 = log2.loc[log2.Comment!="Chile"]
            log4 = log2.loc[log2.Comment=="Chile"]
            plt.scatter(log3["(D-d)/d^2"], log3.Rinfy**0.5/(log3.D-log3.d), color=cmap(num), label="{0:d}-{1:d}".format(bs,bs+binsize))
            plt.scatter(log4["(D-d)/d^2"], log4.Rinfy**0.5/(log4.D-log4.d), edgecolors=cmap(num), marker="^", fc=(0,0,0,0))
        plt.xlabel("$(D-d)/d^2$")
        plt.ylabel("$R_\infty / (D-d)$")
        plt.legend(ncol=2, fontsize=6)
        plt.grid(which="both", ls=":")
        plt.loglog()
    def scatter(self, x="D-d", y="DA_fit", xlabel=None, ylabel=None, ax=None, mode="log", highlight_Chile_data=True):
        """I want to implement a more flexible plotting tool to test ideas.
        The data will still be plotted in OD bins of 20. The args allow one to specify which column(s) to plot,
        and what label(s) to use for the xy-axes.
        Edit:
        05172022 -- Initial commit."""
        log1 = self.data
        # log1 = log.dropna(subset=["Rinfy", "t2"])
        binsize = 10 # OD bin size
        if ax == None:
            fig, ax = plt.subplots(figsize=(3.5,3), dpi=300)
        bin_starts = range(0, int(log1.OD.max()), binsize)
        cmap = plt.cm.get_cmap("tab10")
        for num, bs in enumerate(bin_starts):
            log2 = log1.loc[(log1.OD>bs)&(log1.OD<=bs+binsize)]
            if len(log2) == 0:
                continue
            log3 = log2.loc[log2.Comment!="Chile"]
            log4 = log2.loc[log2.Comment=="Chile"]
            if highlight_Chile_data == True:
                ax.scatter(log3[x], log3[y], color=cmap(num), label="{0:d}-{1:d}".format(bs,bs+binsize))
                ax.scatter(log4[x], log4[y], edgecolors=cmap(num), marker="^", fc=(0,0,0,0))
            else:
                ax.scatter(log2[x], log2[y], color=cmap(num), label="{0:d}-{1:d}".format(bs,bs+binsize))
        if xlabel == None:
            xlabel = x
        if ylabel == None:
            ylabel = y
        ax.set_xlabel("{}".format(xlabel))
        ax.set_ylabel("{}".format(ylabel))
        ax.legend(ncol=1, fontsize=6, loc="upper left")
        if mode == "log":
            ax.grid(which="both", ls=":")
            ax.loglog()
        return ax

# %% codecell
if __name__=="__main__":
    # test de_data class
    # %% codecell
    # make all the plots
    log_dir = r"C:\Users\liuzy\Documents\Github\DE\Data\structured_log_DE.ods"
    log = pd.read_excel(io=log_dir, sheet_name="main")
    data = de_data(log)
    # %% codecell
    data.parameter_space(highlight_Chile_data=True) # 1
    data.plot_MSD_model_Cristian() # 3
    data.scatter_0(mode="log", highlight_Chile_data=True) # 2
    data.plot_0(nbins=5, overlap=0, mode="log") # 4
    data.scatter_1(mode="log", highlight_Chile_data=True) # 5
    data.plot_1(nbins=5, overlap=0, mode="log") # 6
    data.Rinf2_tau() # 7
    data.Rinf2_over_tau() # 8
    data.rescale_Rinf_OD() # 9
    data.rescale_Rinf_freespace() # 10
    # %% codecell
    # generate_msd_repo method generates .jpg images, and take longer time to run
    # (several minutes), use caution when running this block
    data.generate_msd_repo(component="y", data_dir=r"C:\Users\liuzy\Documents\Github\DE\Data\traj")
# %% codecell
class drop_data:
    """Droplet data plotting tool."""
    def __init__(self, data):
        """Initialize a data object with main log spreadsheet (pd.DataFrame)"""
        self.data = data
        self.OD_to_nc = 8e8 # OD to number concentration conversion factor, cells/ml
        self.single_bacterial_volume = 1 # um^3
    def __repr__(self):
        return self.data.__repr__()
    def parameter_space(self):
        """Plot the parameter space of current data, D vs. OD"""
        fig, ax = plt.subplots(figsize=(3.5, 3), dpi=100)
        ax.scatter(self.data["Bacterial concentration"], self.data["Droplet size"])
        ax.set_xlabel("OD")
        ax.set_ylabel("$D$ (um)")
    def find_lifetime_data(self, n=5):
        """Return the Droplet#'s of droplets with more than n videos, n is 5 by default"""
        self.lifetime_data_list = []
        for i in self.data["Droplet#"].drop_duplicates():
            subdata = self.data.loc[self.data["Droplet#"]==i]
            if len(subdata) >= n:
                self.lifetime_data_list.append(i)
        # print("Lifetime data located: {}".format(str(self.lifetime_data_list)))
    def plot_mean_velocity_evolution(self, n=5, mode="log"):
        """Plot mean velocity vs. time.
        Use the time of first video as 0 (or 1 in log mode)
        n: number of curves on each plot
        mode: the scale of time axis, can be 'lin' or 'log' (default)"""
        self.find_lifetime_data()
        cmap = plt.cm.get_cmap("Set1")
        for num, i in enumerate(self.lifetime_data_list):
            if num % n == 0:
                fig, ax = plt.subplots(figsize=(3.5, 2), dpi=100)
                ax.set_xlabel("time (min)")
                ax.set_ylabel("mean velocity (um/s)")
                ax.set_xlim([0, 60])
                if mode == "log":
                    ax.set_xlim([1, 60])
                    ax.set_xscale("log")
                ax.set_ylim([0, 15])
            subdata = self.data.loc[self.data["Droplet#"]==i]
            t = subdata["Time in minutes"]
            t -= subdata["Time in minutes"].min()
            v = subdata["Initial mean velocity (10 s)"]
            if mode == "log":
                t += 1
            ax.plot(t, v, marker="s", color=cmap(num % n), label="{:d}".format(i))
            ax.legend(frameon=False)
    def plot_droplet_size_evolution(self, n=5, mode="log"):
        """Plot droplet size vs. time.
        Use the time of first video as 0 (or 1 in log mode)
        n: number of curves on each plot
        mode: the scale of time axis, can be 'lin' or 'log' (default)"""
        self.find_lifetime_data()
        cmap = plt.cm.get_cmap("Set1")
        for num, i in enumerate(self.lifetime_data_list):
            if num % n == 0:
                fig, ax = plt.subplots(figsize=(3.5, 2), dpi=100)
                ax.set_xlabel("time (min)")
                ax.set_ylabel("droplet diameter (um)")
                ax.set_xlim([0, 60])
                if mode == "log":
                    ax.set_xlim([1, 60])
                    ax.set_xscale("log")
                if mode == "loglog":
                    ax.set_xlim([1, 60])
                    ax.set_xscale("log")
                    ax.set_yscale("log")
                # ax.set_ylim([0, 15])
            subdata = self.data.loc[self.data["Droplet#"]==i]
            t = subdata["Time in minutes"]
            t -= subdata["Time in minutes"].min()
            D = subdata["Droplet size"]
            if mode == "log":
                t += 1
            ax.plot(t, D, marker="s", color=cmap(num % n), label="{:d}".format(i))
            ax.legend(frameon=False)
    def plot_volume_fraction_evolution(self, n=5, mode="log"):
        """Plot droplet size vs. time.
        Use the time of first video as 0 (or 1 in log mode)
        n: number of curves on each plot
        mode: the scale of time axis, can be 'lin' or 'log' (default)"""
        self.find_lifetime_data()
        cmap = plt.cm.get_cmap("Set1")

        for num, i in enumerate(self.lifetime_data_list):
            if num % n == 0:
                fig, ax = plt.subplots(figsize=(3.5, 2), dpi=100)
                ax.set_xlabel("time (min)")
                ax.set_ylabel("volume fraction")
                ax.set_xlim([0, 60])
                if mode == "log":
                    ax.set_xlim([1, 60])
                    ax.set_xscale("log")
                ax.set_ylim([0.1, 0.3])
            subdata = self.data.loc[self.data["Droplet#"]==i]
            initial_droplet_volume = 4/3 * np.pi * (subdata["Droplet size"].iloc[0]/2) ** 3
            bacterial_volume = subdata["Bacterial concentration"].iloc[0] * self.OD_to_nc * initial_droplet_volume * 1e-12 * self.single_bacterial_volume

            t = subdata["Time in minutes"]
            t -= subdata["Time in minutes"].min()
            vf = bacterial_volume / (4/3 * np.pi * (subdata["Droplet size"]/2) ** 3)
            if mode == "log":
                t += 1
            ax.plot(t, vf, marker="s", color=cmap(num % n), label="{:d}".format(i))
            ax.legend(frameon=False)
    def plot_velocity_volume_fraction_correlation(self, time_bins=5):
        """Plot the correlation between mean velocity and volume fraction
        time_bins: number of time bins"""
        self.find_lifetime_data()
        cmap = plt.cm.get_cmap("Set1")
        bin_size = 60 // time_bins
        bin_starts = range(0, 60, bin_size)
        plot_data_list = []
        for num, i in enumerate(self.lifetime_data_list):
            subdata = self.data.loc[self.data["Droplet#"]==i]
            initial_droplet_volume = 4/3 * np.pi * (subdata["Droplet size"].iloc[0]/2) ** 3
            bacterial_volume = subdata["Bacterial concentration"].iloc[0] * self.OD_to_nc * initial_droplet_volume * 1e-12 * self.single_bacterial_volume
            t = subdata["Time in minutes"]
            t -= subdata["Time in minutes"].min()
            vf = bacterial_volume / (4/3 * np.pi * (subdata["Droplet size"]/2) ** 3)
            plot_data = subdata.assign(t=t, vf=vf)
            plot_data_list.append(plot_data)
        data = pd.concat(plot_data_list, axis=0)
        fig, ax = plt.subplots(figsize=(3.5, 3), dpi=100)
        for num, start in enumerate(bin_starts):
            subdata = data.loc[(data.t>start)&(data.t<=start+bin_size)]
            ax.scatter(subdata.vf, subdata["Initial mean velocity (10 s)"],
                       color=cmap(num), label="{0:.0f}-{1:.0f}".format(start, start+bin_size))
        ax.set_xlabel("volume fraction")
        ax.set_ylabel("mean velocity")
        ax.legend()

if __name__=="__main__":
    # %% codecell
    # Test droplet_image class
    # %% codecell
    # create object
    folder = "test_images/moving_mask_piv/raw"
    l = readdata(folder, "tif")
    DI = droplet_image(l, 50, 0.33)
    # %% codecell
    # test __repr__
    DI
    # %% codecell
    # test droplet_traj()
    mask_dir = r"test_images\moving_mask_piv\mask.tif"
    mask = io.imread(mask_dir)
    xy0 = (178, 161)
    traj = DI.droplet_traj(mask, xy0)
    traj
    # %% codecell
    # test check_traj()
    mask_shape = (174, 174)
    DI.check_traj(traj, mask_shape, n=8)
    # %% codecell
    img = DI.get_image(0)
    plt.imshow(img)
    # %% codecell
    mask_shape = (174, 174)
    img = DI.get_cropped_image(0, traj, mask_shape)
    plt.imshow(img)
    # %% codecell
    # test moving_mask_piv()
    save_folder = r"test_images\moving_mask_piv\piv_result"
    winsize = 20
    overlap = 10
    dt = 0.02
    mask_dir = r"test_images\moving_mask_piv\mask.tif"
    xy0 = (178, 161)
    mask_shape = (174, 174)
    DI.moving_mask_piv(save_folder, winsize, overlap, dt, mask_dir, xy0, mask_shape)
    # %% codecell
    # test piv_overlay_moving()
    piv_folder = r"test_images\moving_mask_piv\piv_result"
    out_folder = r"test_images\moving_mask_piv\piv_overlay_moving"
    traj = pd.read_json(os.path.join(piv_folder, "droplet_traj.json"))
    with open(os.path.join(piv_folder, "piv_params.json"), "r") as f:
        piv_params = json.load(f)
    DI.piv_overlay_moving(piv_folder, out_folder, traj, piv_params, sparcity=1)
    # %% codecell
    # output cropped images for note figure
    out_folder = r"test_images\moving_mask_piv\cropped_images"
    if os.path.exists(out_folder) == False:
        os.makedirs(out_folder)
    dpi = 300
    figscale = 1
    for i in DI.sequence.index:
        img = DI.get_cropped_image(i, traj, (174, 174))
        w, h = img.shape[1] / dpi, img.shape[0] / dpi
        fig = Figure(figsize=(w*figscale, h*figscale)) # on some server `plt` is not supported
        canvas = FigureCanvas(fig) # necessary?
        ax = fig.add_axes([0, 0, 1, 1])
        ax.imshow(img, cmap='gray')
        ax.axis("off")
        fig.savefig(os.path.join(out_folder, "{}.jpg".format(DI.get_image_name(i))))
    # %% codecell
    # test fixed_mask_piv()
    winsize = 20
    overlap = 10
    dt = 0.02
    mask_dir = "test_images/moving_mask_piv/mask.tif"
    piv, params = DI.fixed_mask_piv(winsize, overlap, mask_dir)
    params
    # %% codecell
    # test piv_overlay_fixed()
    piv_folder = r"test_images\fixed_mask_piv\piv_result"
    out_folder = r"test_images\fixed_mask_piv\piv_overlay_fixed"
    sparcity = 1
    DI.piv_overlay_fixed(piv_folder, out_folder, sparcity=sparcity)



    # %% codecell
    # test drop_data class
    # create drop_data object
    log_dir = r"..\Data\structured_log.ods"
    log = pd.read_excel(io=log_dir, sheet_name="main")
    dd = drop_data(log)
    dd.parameter_space()
    dd.find_lifetime_data()
    dd.plot_mean_velocity_evolution(n=6, mode="log")
    dd.plot_droplet_size_evolution(n=6, mode="log")
    # %% codecell
    dd.plot_volume_fraction_evolution(n=6, mode="lin")
    # %% codecell
    dd.plot_velocity_volume_fraction_correlation(time_bins=6)



import numpy as np
import matplotlib.pyplot as plt
from myimagelib.myImageLib import dirrec, bpass
from skimage import io, util
import os
import pandas as pd
import sys
import time
from numpy.polynomial.polynomial import polyvander

def corrS(X, Y, U, V):
    """
    Compute the spatial autocorrelations of a velocity field.

    :param X,Y,U,V: the result of PIV analysis. Each is a 2D array. Use :py:func:`pivLib.read_piv` to construct thses arrays from PIV data files.
    :type X,Y,U,V: 2d arrays
    :return: x, y, CA, CV -- The angle autocorrelation (CA) and velocity autocorrelation (CV). x, y are the associated distances.

    .. note::

        We only consider half of the length in each dimension, so x, y are different from the input X, Y.

    .. rubric:: Edit

    * Dec 13, 2021 -- i) Replace all the `mean()` function to nanmean, to handle masked PIV data. ii) Add doc string.
    * Dec 15, 2021 -- if norm `vsqrt` is 0, set it to np.nan to avoid divided by zero warning!
    * Dec 16, 2021 -- Shift the output X, Y origin to 0, 0, so that 0 distance will have the correlation function = 1. More intuitive.
    * Jan 05, 2023 -- Adapt myimagelib import style.
    """
    row, col = X.shape
    r = int(row/2)
    c = int(col/2)
    vsqrt = (U ** 2 + V ** 2) ** 0.5
    vsqrt[vsqrt==0] = np.nan # if norm is 0, set it to np.nan to avoid divided by zero warning!
    U = U - np.nanmean(U)
    V = V - np.nanmean(V)
    Ax = U / vsqrt
    Ay = V / vsqrt
    CA = np.ones((r, c))
    CV = np.ones((r, c))
    # pdb.set_trace()
    for xin in range(0, c):
        for yin in range(0, r):
            if xin != 0 or yin != 0:
                CA[yin, xin] = np.nanmean((Ax[0:row-yin, 0:col-xin] * Ax[yin:row, xin:col] + Ay[0:row-yin, 0:col-xin] * Ay[yin:row, xin:col]))
                CV[yin, xin] = np.nanmean((U[0:row-yin, 0:col-xin] * U[yin:row, xin:col] + V[0:row-yin, 0:col-xin] * V[yin:row, xin:col])) / (np.nanstd(U)**2+np.nanstd(V)**2)
    return X[0:r, 0:c] - X[0, 0], Y[0:r, 0:c] - Y[0, 0], CA, CV

def corrI(X, Y, I):
    """
    Compute pixel intensity spatial autocorrelation of an image.

    :param X: x coordinates of image pixels
    :type X: 2d array
    :param Y: y coordinates of image pixels
    :type Y: 2d array
    :param I: image
    :type I: 2d array
    :return: XI, YI, CI, where CI is the autocorrelation map and XI and YI are the corresponding coordinates.
    :rtype: 2d arrays
    """
    row, col = I.shape
    I = I - I.mean()
    # CI = np.ones(I.shape)
    r = int(row/2)
    c = int(col/2)
    CI = np.ones((r, c))
    XI = X[0: r, 0: c]
    YI = Y[0: r, 0: c]
    normalizer = I.std() ** 2
    for xin in range(0, int(col/2)):
        for yin in range(0, int(row/2)):
            if xin != 0 or yin != 0:
                I_shift_x = np.roll(I, xin, axis=1)
                I_shift = np.roll(I_shift_x, yin, axis=0)
                CI[yin, xin] = (I[yin:, xin:] * I_shift[yin:, xin:]).mean() / normalizer
    return XI, YI, CI


def divide_windows(img, windowsize=[20, 20], step=10):
    """
    Divide an image into windows with specified size and step. Size and step can be different, so that we generate a downsampled image with either overlap or gaps.

    :param img: image
    :type img: 2d array
    :param windowsize: window size in x and y
    :type windowsize: list(int)
    :param step: distance between the centers of adjacent windows
    :type step: int
    :return: X,Y,I -- I is the downsampled image, while X, Y are the corresponding coordinates in x an y
    :rtype: 2d arrays
    """
    row, col = img.shape
    windowsize[0] = int(windowsize[0])
    windowsize[1] = int(windowsize[1])
    step = int(step)
    if isinstance(windowsize, list):
        windowsize = tuple(windowsize)
    X = np.array(range(0, col-windowsize[0], step))
    Y = np.array(range(0, row-windowsize[1], step))
    X, Y = np.meshgrid(X, Y)
    I = util.view_as_windows(img, windowsize, step=step).mean(axis=(2, 3))
    return X, Y, I


def distance_corr(X, Y, C):
    """ Convert 2d correlation matrix into 1d.

    :param X,Y: 2d coordinates, can be either 2d matrices or flattened matrices
    :param C: correlation matrix, can be either 2d matrices or flattened matrices
    :return: a DataFrame of ["R", "C"].

    .. rubric:: Edit

    :12162021:
     * check input dimension, if it's 1, do not use `flatten`,
     * add doc string
    """
    if len(X.shape) == 2:
        r_corr = pd.DataFrame({'R': (X.flatten()**2 + Y.flatten()**2) ** 0.5, 'C': C.flatten()}).sort_values(by='R')
    else:
        r_corr = pd.DataFrame({'R': (X**2 + Y**2) ** 0.5, 'C': C}).sort_values(by='R')
    return r_corr



def density_fluctuation(img8):
    """
    This is the first attempt to calculate density fluctuations in bacterial suspensions. There exists 2 methods: i) compute the spatial variation in each image, with respect to various box size, then average over time; ii) compute temporal variation in at different spot, with respect to various box size, then average over space.

    Method implemented in this function is the first one. It turned out that even for a static video, we can easily measure a very strong density fluctuation, which cannot be true. We later realize that, this is due to the intrinsic spatial light variation of the bright field microscopy.

    One way to solve this problem is to subtract the intrinsic variation from the raw images. However, at that time we did not succeed in doing it (as I am writing this document, I konw better how to do it). So we chose the second way: to use method (ii) for the density fluctuation calculation. Different from method (i), method (ii) is not sensitive to the spatial variation in each image, but considers how the intensities evolve over time. Therefore, we can get density fluctuations that make sense.

    .. warning::

        Currently, we deprecate this method. It is not used in our publication *Soft Matter* **17**: 10806–17. https://doi.org/10.1039/D1SM01183A.


    :param img8: 8-bit grayscale image
    :type img8: 2d array
    :return: density fluctuation data, a 2-column table containing the box size (area) "n" and the intensity standard deviation over time "d"

    """
    row, col = img8.shape
    l = min(row, col)
    size_min = 5
    boxsize = np.unique(np.floor(np.logspace(np.log10(size_min), np.log10((l-size_min)/2), 100)))
    # Gradually increase box size and calculate dN=std(I) and N=mean(I)
    # choose maximal box size to be (l-size_min)/2
    # to guarantee we have multiple boxes for each calculation, so that
    # the statistical quantities are meaningful.
    # Step is chosen as 5*size_min to guarantee speed as well as good statistics
    # instead of box size. When box size is large, number of boxes is too small
    # to get good statistics.
    # Papers by Narayan et al. used different method to calculate density fluctuation
    # Igor Aranson commented on the averaging methods saying that in a spatially
    # homogeneous system (small spatial temporal correlation) two methods should match.
    # This suggests that I need to test both methods.

    NList = []
    dNList = []
    for bs in boxsize:
        X, Y, I = divide_windows(img8, windowsize=[bs, bs], step=bs)
        N = bs*bs
        # dN = np.log10(I).std()*bs*bs
        dN = I.std()*bs*bs
        NList.append(N)
        dNList.append(dN)
    df_data = pd.DataFrame().assign(n=NList, d=dNList)
    return df_data


def div_field(img, pivData, winsize, step):
    """
    Compute multiple divergence fields from PIV data and microscopy image.

    :param img: 2d array
    :param pivData: DataFrame with columns (x, y, u, v)
    :param winsize: window size of PIV data
    :param step: step size of PIV data

    .. note::

        :code:`winsize` and :code:`step` should be consistent with the parameters of pivData, be extra coutious!

    :return: divergence fields of given data. (c, v, divcn, divcv, divv). *c* is concentration inferred from the image.


    """
    X, Y, I = divide_windows(img, windowsize=[winsize, winsize], step=step)
    # concentration field
    I0 = 255
    c = I0 - I

    # calculation for divcn and divcv
    row, col = I.shape
    vx = np.array(pivData.u).reshape(I.shape)
    vy = np.array(pivData.v).reshape(I.shape)
    v = (vx**2 + vy**2)**.5
    nx = np.array(pivData.u / (pivData.u**2 + pivData.v**2)**.5).reshape(I.shape)
    ny = np.array(pivData.v / (pivData.u**2 + pivData.v**2)**.5).reshape(I.shape)
    cnx = c * nx
    cny = c * ny
    cvx = c * vx
    cvy = c * vy
    divcn = np.zeros(I.shape)
    divcv = np.zeros(I.shape)
    divv = np.zeros(I.shape)
    for x in range(0, col-1):
        for y in range(0, row-1):
            divcn[y, x] = cnx[y,x+1] - cnx[y,x] + cny[y+1,x] - cny[y,x]
            divcv[y, x] = cvx[y,x+1] - cvx[y,x] + cvy[y+1,x] - cvy[y,x]
            divv[y, x] = vx[y,x+1] - vx[y,x] + vy[y+1,x] - vy[y,x]
    return c, v, divcn, divcv, divv


def df2(imgstack, size_min=5, step=250, method='linear'):
    """
    This is used for small scale test of temporal variation based density fluctuation analysis.
    Here the input is a 3 dimensional array that contains all the images in a video [stack_index, height, width].
    For larger scale application, use the script df2.py or a later implementation of df2 which can take directory as argument 1.

    :param imgstack: 3-D array with dimensions [stack_index, height, width]
    :param size_min: minimal box size to sample
    :param step: distance between adjacent boxes in a frame
    :param method: use pixel intensity directly as concentration ('linear'), or use the log of it ('log')

    :return: the spatial average of temporal variations
    """

    L = min(imgstack.shape[1:3])
    boxsize = np.unique(np.floor(np.logspace(np.log10(size_min),
                        np.log10((L-size_min)/2),100)))

    df = pd.DataFrame()
    for i, img in enumerate(imgstack):
        framedf = pd.DataFrame()
        for bs in boxsize:
            X, Y, I = divide_windows(img, windowsize=[bs, bs], step=step)
            tempdf = pd.DataFrame().assign(I=I.flatten(), t=int(i), size=bs,
                           number=range(0, len(I.flatten())))
            framedf = framedf.append(tempdf)
        df = df.append(framedf)

    if method == 'log':
        df['I'] = np.log(df['I'])

    df_out = pd.DataFrame()
    for number in df.number.drop_duplicates():
        subdata1 = df.loc[df.number==number]
        for s in subdata1['size'].drop_duplicates():
            subdata = subdata1.loc[subdata1['size']==s]

            d = s**2 * np.array(subdata.I).std()
            n = s**2
            tempdf = pd.DataFrame().assign(n=[n], d=d, size=s, number=number)
            df_out = df_out.append(tempdf)

    average = pd.DataFrame()
    for s in df_out['size'].drop_duplicates():
        subdata = df_out.loc[df_out['size']==s]
        avg = subdata.drop(columns=['size', 'number']).mean().to_frame().T
        average = average.append(avg)

    return average

def df2_(img_stack, boxsize=None, size_min=5, step=250):
    """
    Compute number fluctuations of an image stack (3D np.array, frame*h*w)

    :param img_stack: a stack of image, a 3D array
    :param boxsize: a list-like of integers, specify the subsystem size to look at. If None, generate a log spaced array within (size_min, L/2) (pixels), with 100 points
    :param size_min: the smallest box size
    :param step: step used when dividing image into windows

    :return: DataFrame of n and d, where n is box area (px^2) and d is total number fluctuations (box_area*dI)
    """
    L = min(img_stack.shape[1:3])
    if boxsize == None:
        boxsize = np.unique(np.floor(np.logspace(np.log10(size_min), np.log10(L/2), 100)))

    dI_list = []
    for bs in boxsize:
        I = divide_stack(img_stack, winsize=[bs, bs], step=step)
        dI = I.std(axis=0).mean() * bs ** 2
        dI_list.append(dI)

    return pd.DataFrame({'n': np.array(boxsize)**2, 'd': dI_list})

def divide_stack(img_stack, winsize=[50, 50], step=25):
    """
    Divide image stack into several evenly spaced windows of stack. Average the pixel intensity within each window. For example, a 30*30*30 image stack, if we apply a divide_stack, with winsize=[15, 15] and step=15, the resulting divided stack will be (30, 4).

    :param img_stack: a stack of image, a 3D array, axis0 should be frame number
    :param winsize: division scheme, default to [50, 50]
    :param step: division scheme, default ot 25

    :return: the result
    """
    length = img_stack.shape[0]
    divide = util.view_as_windows(img_stack, window_shape=[length, *winsize], step=step).mean(axis=(-1, -2))
    # reshape
    divided_array = divide.reshape((np.prod(divide.shape[:3]), length)).transpose()

    return divided_array

def plot_gnf(gnf_data):
    """
    Used for small scale test of gnf analysis. It incorporates the guide of the eye slope already in the function, which usually needs further adjustment when preparing paper figures.

    :param gnf_data: gnf data generated by df2()

    :return ax: axis on which the data are plotted
    :return slope: the slope of the gnf_data curve
    """
    x = gnf_data.n / 100
    y = gnf_data.d / x ** 0.5
    y = y / y.iat[0]
    fig = plt.figure()
    ax = fig.add_axes([0,0,1,1])
    ax.plot(x, y)
    xf, yf, xt, yt, slope = label_slope(x, y, location='n')
    ax.plot(xf, yf, ls='--', color='black')
    ax.text(xt, yt, '{:.2f}'.format(slope))
    ax.loglog()
    ax.set_xlabel('$l^2/l_b^2$')
    ax.set_ylabel('$\Delta N/\sqrt{N}$')
    return ax, slope

def vorticity(pivData, step=None, shape=None):
    """
    Compute vorticity field based on piv data (x, y, u, v)

    :param pivData: DataFrame of (x, y, u, v)
    :param step: distance (pixel) between adjacent PIV vectors

    :return: vorticity field of the velocity field. unit: [u]/pixel, [u] is the unit of u, usually px/s
    """
    x = pivData.sort_values(by=['x']).x.drop_duplicates()
    if step == None:
        # Need to infer the step size from pivData
        step = x.iat[1] - x.iat[0]

    if shape == None:
        # Need to infer shape from pivData
        y = pivData.y.drop_duplicates()
        shape = (len(y), len(x))

    X = np.array(pivData.x).reshape(shape)
    Y = np.array(pivData.y).reshape(shape)
    U = np.array(pivData.u).reshape(shape)
    V = np.array(pivData.v).reshape(shape)

    dudy = np.gradient(U, step, axis=0)
    dvdx = np.gradient(V, step, axis=1)
    vort = dvdx - dudy

    return vort

def convection(pivData, image, winsize, step=None, shape=None):
    """
    Compute convection term u.grad(c) based on piv data (x, y, u, v) and image.

    :param pivData: -- DataFrame of (x, y, u, v)
    :param image: -- the image corresponding to pivData
    :param winsize: -- coarse-graining scheme of image
    :param step: -- (optional) distance (pixel) between adjacent PIV vectors
    :param shape: -- (optional) shape of piv matrices

    :return: convection term u.grad(c). unit: [u][c]/pixel, [u] is the unit of u, usually px/s, [c] is the unit of concentration measured from image intensity, arbitrary.
    """
    x = pivData.sort_values(by=['x']).x.drop_duplicates()
    if step == None:
        # Need to infer the step size from pivData
        step = x.iat[1] - x.iat[0]

    if shape == None:
        # Need to infer shape from pivData
        y = pivData.y.drop_duplicates()
        shape = (len(y), len(x))

    # check coarse-grained image shape
    X, Y, I = divide_windows(image, windowsize=[winsize, winsize], step=step)
    assert(I.shape==shape)

    X = np.array(pivData.x).reshape(shape)
    Y = np.array(pivData.y).reshape(shape)
    U = np.array(pivData.u).reshape(shape)
    V = np.array(pivData.v).reshape(shape)

    # compute gradient of concentration
    # NOTE: concentration is negatively correlated with intensity.
    # When computing gradient of concentration, the shifting direction should reverse.

    dcx = np.gradient(I, -step, axis=1)
    dcy = np.gradient(I, -step, axis=0)

    udc = U * dcx + V * dcy

    return udc

def divergence(pivData, step=None, shape=None):
    """
    Compute divergence field based on piv data (x, y, u, v)

    :param pivData: DataFrame of (x, y, u, v)
    :param step: distance (pixel) between adjacent PIV vectors

    :return: vorticity field of the velocity field. unit: [u]/pixel, [u] is the unit of u, usually px/s
    """
    x = pivData.sort_values(by=['x']).x.drop_duplicates()
    if step == None:
        # Need to infer the step size from pivData
        step = x.iat[1] - x.iat[0]

    if shape == None:
        # Need to infer shape from pivData
        y = pivData.y.drop_duplicates()
        shape = (len(y), len(x))

    X = np.array(pivData.x).reshape(shape)
    Y = np.array(pivData.y).reshape(shape)
    U = np.array(pivData.u).reshape(shape)
    V = np.array(pivData.v).reshape(shape)

    dudx = np.gradient(U, step, axis=1)
    dvdy = np.gradient(V, step, axis=0)
    div = dudx + dvdy

    return div

def local_df(img_folder, seg_length=50, winsize=50, step=25):
    """
    Compute local density fluctuations of given image sequence in img_folder

    :param img_folder: folder containing .tif image sequence
    :param seg_length: number of frames of each segment of video, for evaluating standard deviations
    :param winsize: window size
    :param step: step between adjacent windows

    :return: dict containing 't' and 'local_df', 't' is a list of time (frame), 'std' is a list of 2d array with local standard deviations corresponding to 't'
    """

    l = readseq(img_folder)
    num_frames = len(l)
    assert(num_frames>seg_length)

    stdL = []
    tL = range(0, num_frames, seg_length)
    for n in tL:
        img_nums = range(n, min(n+seg_length, num_frames))
        l_sub = l.loc[img_nums]
        img_seq = []
        for num, i in l_sub.iterrows():
            img = io.imread(i.Dir)
            X, Y, I = divide_windows(img, windowsize=[50, 50], step=25)
            img_seq.append(I)
        img_stack = np.stack([img_seq], axis=0)
        img_stack = np.squeeze(img_stack)
        std = np.std(img_stack, axis=0)
        stdL.append(std)

    return {'t': tL, 'std': stdL}

def compute_energy_density(pivData, d=25*0.33, MPP=0.33):
    """
    Compute kinetic energy density in k space from piv data. The unit of the return value is [velocity] * [length], where [velocity] is the unit of pivData, and [length] is the unit of sample_spacing parameter. Note, the default value of sampling_spacing does not have any significance. It is just the most convenient value for my first application, and should be set with caution when a different magnification and PIV are used.

    :param pivData: 2D piv data, DataFrame of x, y, u, v
    :param d: sample spacing, in unit of microns
    :param MPP: microns per pixel

    :return: kinetic energy field in wavenumber (k) space

    .. rubric:: Test

    >>> pivData = pd.read_csv(r'E:\\moreData\\08032020\piv_imseq\\01\\3370-3371.csv')
    >>> compute_energy_density(pivData)

    .. rubric:: Edit

    :11020202:
        Add parameter sample_spacing, the distance between adjacent velocity (or data in general. The spacing is used to rescale the DFT so that it has a unit of [velocity] * [length]. In numpy.fft, the standard fft function is defined as

        .. math::

            A_k = \\sum\\limits^{n-1}_{m=0} a_m \\exp \\left[ -2\pi i \\frac{mk}{n} \\right]

        The unit of this Fourier Transform :math:`A_k` is clearly the same as :math:`a_m`. In order to get unit [velocity] * [length], and to make transform result consistent at different data density, I introduce sample_spacing :math:`d` as a modifier of the DFT. After this modification, the energy spectrum computed at various step size (of PIV) should give quantitatively similar results.

    :11042020:
        Replace the (* sample_spacing * sample_spacing) after v_fft with ( / row / col). This overwrites the edit I did on 11022020.

    :11112020: removed ( / row / col), add the area constant in energy_spectrum() function. Details can be found `here <https://zloverty.github.io/research/DF/blogs/energy_spectrum_2_methods_11112020.html>`_

    :11302020:
        *  Convert unit of velocity, add optional arg MPP (microns per pixel);
        * Add * d * d to both velocity FFT's to account for the missing length in default FFT algorithm;
        * The energy spectrum calculated by this function shows a factor of ~3 difference when comparing \int E(k) dk with v**2.sum()/2
    """

    row = len(pivData.y.drop_duplicates())
    col = len(pivData.x.drop_duplicates())
    U = np.array(pivData.u).reshape((row, col)) * MPP
    V = np.array(pivData.v).reshape((row, col)) * MPP

    u_fft = np.fft.fft2(U) * d * d
    v_fft = np.fft.fft2(V) * d * d

    E = (u_fft * u_fft.conjugate() + v_fft * v_fft.conjugate()) / 2

    return E

def compute_wavenumber_field(shape, d):
    """
    Compute the wave number field Kx and Ky, and magnitude field k.
    Note that this function works for even higher dimensional shape.

    :param shape: shape of the velocity field and velocity fft field, tuple
    :param d: sample spacing. This is the distance between adjacent samples, for example, velocities in PIV. The resulting frequency space has the unit which is inverse of the unit of d. The preferred unit of d is um.

    :return:
        * k -- wavenumber magnitude field
        * K -- wavenumber fields in given dimensions

    .. rubric:: Test

    >>> shape = (5, 5)
    >>> k, K = compute_wavenumber_field(shape, 0.2)

    .. rubric:: Edit

    :12022020: multiply 2pi to the wavenumber to account for the built-in 2pi in the fft method. This factor leads to a difference in the magnitude of 1D energy spectra.

        .. note:: The dimensionless wavenumber should remain unchanged.
    """

    for num, length in enumerate(shape):
        kx = np.fft.fftfreq(length, d=d)
        if num == 0:
            k = (kx,)
        else:
            k += (kx,)

    K = np.meshgrid(*k, indexing='ij')

    for num, k1 in enumerate(K):
        if num == 0:
            ksq = k1 ** 2
        else:
            ksq += k1 ** 2

    k_mag = ksq ** 0.5 * 2 * np.pi

    return k_mag, K

def energy_spectrum(pivData, d=25*0.33):
    """
    Compute energy spectrum (E vs k) from pivData.


    :param pivData: piv data
    :param d: sample spacing. This is the distance between adjacent samples, for example, velocities in PIV. The resulting frequency space has the unit which is inverse of the unit of d. The default unit of d is um.

    :return: energy spectrum, DataFrame (k, E)

    .. rubric:: Edit

    :10192020: add argument d as sample spacing
    :11112020: add area constant, see details `here <https://zloverty.github.io/research/DF/blogs/energy_spectrum_2_methods_11112020.html>`_
    :11302020: The energy spectrum calculated by this function shows a factor of ~3 difference when comparing \int E(k) dk with v**2.sum()/2
    """

    row = len(pivData.y.drop_duplicates())
    col = len(pivData.x.drop_duplicates())

    E = compute_energy_density(pivData, d) / (row * d * col * d)
    k, K = compute_wavenumber_field(E.shape, d)

    ind = np.argsort(k.flatten())
    k_plot = k.flatten()[ind]
    E_plot = E.real.flatten()[ind]

    es = pd.DataFrame(data={'k': k_plot, 'E': E_plot})

    return es

def autocorr1d(x, t):
    """Compute the temporal autocorrelation of a 1-D signal.

    :param x: 1-D signal,
    :param t: the corresponding time of the signal, should be np.array 1d
    :return:
        * corr -- correlation array, with lag time as index
        * lagt -- lag time of the correlation function

    .. rubric:: Edit

    :07272022: Handle time series with missing values.
    """
    if any(np.isnan(x)):
        xn = x - np.nanmean(x)
        x2 = np.nanmean(xn * xn)
        c_list = []
        for s in range(0, len(xn)//2):
            xx = np.roll(xn, shift=s)
            xx[:s] = np.nan
            c_list.append(np.nanmean(xn * xx ))
        corr = np.array(c_list) / x2
    else:
        xn = x - np.nanmean(x)
        corr = np.correlate(xn, xn, mode="same")[len(xn)//2:] / np.inner(xn, xn)
    lagt = (t - t[0])[:len(corr)]
    return corr, lagt

def vacf_piv(vstack, dt, mode="direct"):
    """
    Compute averaged vacf from PIV data. This is a wrapper of function autocorr_t(), adding the averaging over all the velocity spots.

    :param vstack: a 2-D np array of velocity data. axis-0 is time and axes-1,2 are spots in velocity field. Usually, this stack can be constracted by `np.stack(u_list)` and then reshape to flatten axes 1 and 2.
    :param dt: time between two data points
    :param mode: the averaging method, can be "direct" or "weighted". "weighted" will use mean velocity as the averaging weight, whereas "direct" uses 1.
    :return: DataFrame of (corr, t)

    .. rubric:: Edit

    :10262022: add condition :code:`x.sum() != 0`, avoids nan in all-zero columns.
    """
    # rearrange vstack
    assert(len(vstack.shape)==3)
    stack_r = vstack.reshape((vstack.shape[0], -1)).T
    # compute autocorrelation
    corr_list = []
    weight = 1
    normalizer = 0
    for x in stack_r:
        if np.isnan(x.sum()) == False and x.sum() != 0: # masked out part has velocity as nan, which cannot be used for correlation computation
            if mode == "weighted":
                weight = abs(x).mean()
            normalizer += weight
            corr = autocorr1d(x) * weight
            corr_list.append(corr)
            print(x.sum(), corr.sum())
    corr_mean = np.stack(corr_list, axis=0).sum(axis=0) / normalizer

    return pd.DataFrame({"c": corr_mean, "t": np.arange(len(corr_mean)) * dt}).set_index("t")

if __name__ == '__main__':
    img = io.imread(r'I:\Github\Python\Correlation\test_images\GNF\stat\40-1.tif')
    df_data = density_fluctuation(img)



import numpy as np
import matplotlib.pyplot as plt
from scipy.signal import medfilt2d, convolve2d, fftconvolve
from math import exp
import pdb

def normxcorr2(template, image, mode="full"):
    """
    Input arrays should be floating point numbers.
    :param template: N-D array, of template or filter you are using for cross-correlation.
    Must be less or equal dimensions to image.
    Length of each dimension must be less than length of image.
    :param image: N-D array
    :param mode: Options, "full", "valid", "same"
    full (Default): The output of fftconvolve is the full discrete linear convolution of the inputs.
    Output size will be image size + 1/2 template size in each dimension.
    valid: The output consists only of those elements that do not rely on the zero-padding.
    same: The output is the same size as image, centered with respect to the ‘full’ output.
    :return: N-D array of same dimensions as image. Size depends on mode parameter.
    """

    # If this happens, it is probably a mistake
    if np.ndim(template) > np.ndim(image) or \
            len([i for i in range(np.ndim(template)) if template.shape[i] > image.shape[i]]) > 0:
        print("normxcorr2: TEMPLATE larger than IMG. Arguments may be swapped.")

    template = template - np.mean(template)
    image = image - np.mean(image)

    a1 = np.ones(template.shape)
    # Faster to flip up down and left right then use fftconvolve instead of scipy's correlate
    ar = np.flipud(np.fliplr(template))
    out = fftconvolve(image, ar.conj(), mode=mode)

    image = fftconvolve(np.square(image), a1, mode=mode) - \
            np.square(fftconvolve(image, a1, mode=mode)) / (np.prod(template.shape))

    # Remove small machine precision errors after subtraction
    image[np.where(image < 0)] = 0

    template = np.sum(np.square(template))
    out = out / np.sqrt(image * template)

    # Remove any divisions by 0 or very close to 0
    out[np.where(np.logical_not(np.isfinite(out)))] = 0

    return out

def matlab_style_gauss2D(shape=(3,3),sigma=0.5):
    """
    2D gaussian mask - should give the same result as MATLAB's
    fspecial('gaussian',[shape],[sigma])
    """
    m,n = [(ss-1.)/2. for ss in shape]
    y,x = np.ogrid[-m:m+1,-n:n+1]
    h = np.exp( -(x*x + y*y) / (2.*sigma*sigma) )
    h[ h < np.finfo(h.dtype).eps*h.max() ] = 0
    sumh = h.sum()
    if sumh != 0:
        h /= sumh
    return h

def FastPeakFind(data):
    if str(data.dtype) != 'float32':
        data = data.astype('float32')
    # fig = plt.figure()
    # plt.subplot(2,2,1)
    # plt.imshow(data)
    # plt.title('Raw xCorr')
    # pdb.set_trace()

    # plt.subplot(2,2,2)
    mf = medfilt2d(data, kernel_size=3)
    mf = mf.astype('float32')
    # plt.imshow(mf)
    # plt.title('Median filtered')

    # plt.subplot(2,2,3)
    thres = max(min(np.amax(mf,axis=0)), min(np.amax(mf,axis=1)))

    # pdb.set_trace()
    # bw = mf[1 if mf_ > thres for mf_ in mf]

    filt = matlab_style_gauss2D()
    conv = convolve2d(mf, filt, mode='same')
    w_idx = conv > thres
    bw = conv.copy()
    bw[w_idx] = 1
    bw[~w_idx] = 0
    thresholded = np.multiply(bw, conv)
    # plt.imshow(thresholded)
    # plt.title('Thresholded')
    edg = 3
    shape = data.shape
    idx = np.nonzero(thresholded[edg-1: shape[0]-edg-1, edg-1: shape[1]-edg-1])
    idx = np.transpose(idx)
    cent = []
    for xy in idx:
        x = xy[0]
        y = xy[1]
        if thresholded[x, y] >= thresholded[x-1, y-1] and \
            thresholded[x, y] > thresholded[x-1, y] and \
            thresholded[x, y] >= thresholded[x-1, y+1] and \
            thresholded[x, y] > thresholded[x, y-1] and \
            thresholded[x, y] > thresholded[x, y+1] and \
            thresholded[x, y] >= thresholded[x+1, y-1] and \
            thresholded[x, y] > thresholded[x+1, y] and \
            thresholded[x, y] >= thresholded[x+1, y+1]:
            cent.append(xy)
    # for xy in cent:
        # y = xy[0]
        # x = xy[1]
        # plt.plot(x, y, 'ro', markersize=3)
    # plt.show()
    cent = np.asarray(cent).transpose()
    return cent

def maxk(array, num_max):
    array = np.asarray(array)
    length = array.size
    array = array.reshape((1, length))
    idx = np.argsort(array)
    idx2 = np.flip(idx)
    return idx2[0, 0:num_max]

def gauss1(x,a,x0,sigma):
    return a*exp(-(x-x0)**2/(2*sigma**2))


if __name__ == '__main__':
    # data = np.loadtxt('test_array.dat', delimiter='\t')
    # FastPeakFind(data)
    array = np.random.random((1, 10))
    num_max = 2
    ind = maxk(array, num_max)



from openpiv import pyprocess, validation, filters, scaling
from openpiv.smoothn import smoothn
import numpy as np
import matplotlib.pyplot as plt
from skimage import io
import pandas as pd
from scipy.signal import medfilt2d
from myimagelib.myImageLib import readdata, xy_bin
from myimagelib.corrLib import divide_windows, autocorr1d, corrS, distance_corr
import os
import scipy
from scipy.io import savemat
# %% codecell
def PIV1(I0, I1, winsize, overlap, dt, smooth=True):
    """
    Wrapper of :py:func:`openpiv.pyprocess.extended_search_area_piv`, with an option to smooth the resulting velocity field.

    .. deprecated:: 1.0

        This function does not have the frequently required validation and outlier replacing routine, and is replaced permanently by :py:func:`pivLib.PIV`.
    """
    u0, v0 = pyprocess.extended_search_area_piv(I0.astype(np.int32), I1.astype(np.int32), window_size=winsize, overlap=overlap, dt=dt, search_area_size=winsize)
    x, y = pyprocess.get_coordinates(image_size=I0.shape, search_area_size=winsize, window_size=winsize, overlap=overlap)
    if smooth == True:
        u1 = smoothn(u0)[0]
        v1 = smoothn(v0)[0]
        frame_data = pd.DataFrame(data=np.array([x.flatten(), y.flatten(), u1.flatten(), v1.flatten()]).T, columns=['x', 'y', 'u', 'v'])
    else:
        frame_data = pd.DataFrame(data=np.array([x.flatten(), y.flatten(), u0.flatten(), v0.flatten()]).T, columns=['x', 'y', 'u', 'v'])
    return frame_data

def read_piv(pivDir):
    """
    Read piv data from pivDir as X, Y, U, V. pivDir contains *\*.csv* files, which store PIV data of each image pair in a separated file. The data are organized in 4-column tables, (x, y, u, v). This function reconstructs the 2D data, by inferring the dimensions from data.

    :param pivDir: directory of the folder hosting PIV data.
    :return: X, Y, U, V -- 2D PIV data.

    .. rubric:: TEST

    >>> X, Y, U, V = read_piv(pivDir)

    .. rubric:: Edit

    :Nov 17, 2022: Separate functions in two parts, and implement :py:func:`to_matrix`.
    """
    pivData = pd.read_csv(pivDir)

    return to_matrix(pivData)

def to_matrix(pivData):
    """
    Convert PIV data from DataFrame of (x, y, u, v) to four 2D matrices x, y, u, v.

    :param pivData: PIV data
    :type pivData: pandas.DataFrame
    :return: x, y, u, v -- 2D matrices

    .. rubric:: Edit

    :Nov 17, 2022: Initial commit.
    """
    row = len(pivData.y.drop_duplicates())
    col = len(pivData.x.drop_duplicates())
    X = np.array(pivData.x).reshape((row, col))
    Y = np.array(pivData.y).reshape((row, col))
    U = np.array(pivData.u).reshape((row, col))
    V = np.array(pivData.v).reshape((row, col))
    return X, Y, U, V

def PIV(I0, I1, winsize, overlap, dt):
    """
    Standard PIV, consisting of replacing outliers, validate signal to noise ratio, and a smoothing with median filter of kernal shape (3, 3).

    :param I0: The first image
    :param I1: The second iamge
    :param winsize: interrogation window size
    :param step: distance between two windows, usually set to half of window size
    :param dt: time interval between two images, in seconds
    :return:
        * x, y -- coordinates of windows
        * u, v -- velocities in each window

    .. rubric:: Edit

    * Nov 17, 2022 -- (i) Turn off median filter. If needed, do it on the outcome, outside this function. (ii) Update the use of :py:func:`openpiv.pyprocess.get_coordinates`.
    * Dec 06, 2022 -- Update syntax per the changes in the openpiv module. Copy from tutorial.
    * Jan 12, 2023 -- Change s2n ration threshold to 1.
    """
    u0, v0, sig2noise = pyprocess.extended_search_area_piv(
        I0.astype(np.int32),
        I1.astype(np.int32),
        window_size=winsize,
        overlap=overlap,
        dt=dt,
        search_area_size=winsize,
        sig2noise_method='peak2peak',
    )
    # get x, y
    x, y = pyprocess.get_coordinates(
        image_size=I0.shape,
        search_area_size=winsize,
        overlap=overlap
    )
    invalid_mask = validation.sig2noise_val(
        sig2noise,
        threshold = 1.0,
    )
    # replace_outliers
    u2, v2 = filters.replace_outliers(
        u0, v0,
        invalid_mask,
        method='localmean',
        max_iter=3,
        kernel_size=3,
    )
    # median filter smoothing
    # u3 = medfilt2d(u2, 3)
    # v3 = medfilt2d(v2, 3)
    return x, y, u2, v2

def PIV_masked(I0, I1, winsize, overlap, dt, mask):
    """Apply PIV analysis on masked images

    :param I0: the first image
    :param I1: the second image
    :param winsize: same as :py:func:`pivLib.PIV`
    :param overlap: same as :py:func:`pivLib.PIV`
    :param dt: same as :py:func:`pivLib.PIV`
    :param mask: a binary image, that will be convert to a boolean array to mask on PIV data. False marks masked region and True marks the region of interest.
    :return: x, y, u, v -- DataFrame, here x, y is wrt original image, (u, v) are in px/s

    This function is rewritten based on the :py:func:`PIV_droplet()` function in ``piv_droplet.py`` script.
    The intended usage is just to pass one additional `mask` parameter, on top of conventional parameter set.

    .. rubric:: EDIT

    :12142021: Initial commit.
    :12152021:
        * After testing 2 masking procedure, option 1 is better.
        * Two procedures produce similar results, but option 1 is faster.
        * So this function temporarily uses option 1, until a better procedure comes.
    :01072022: Change mask threshold from 1 to 0.5, this will include more velocities.

    .. rubric:: Masking procedure

    * Option 1:
        i) Mask on raw image: I * mask, perform PIV
        ii) Divide mask into windows: mask_w
        iii) use mask_w to mask resulting velocity field: u[~mask_w] = np.nan

    * Option 2:
        i) Perform PIV on raw images
        ii) Divide mask into windows:mask_w
        iii) use mask_w to mask resulting velocity field: u[~mask_w] = np.nan

    .. note::

        This method is specific to ``openpiv`` PIV data. In other PIV methods, the x, y coordinates of windows may be different, and the shape of downsampled mask by :py:func:`corrLib.divide_windows` may not match the PIV results from other methods. Therefore, this method should not be used in the future. Consider to use :py:func:`pivLib.PIV` together with :py:func:`pivLib.apply_mask`.
    """
    assert(mask.shape==I0.shape)
    mask = mask >= mask.mean() # convert mask to boolean array
    I0_mask = I0 * mask
    I1_mask = I1 * mask
    x, y, u, v = PIV(I0_mask, I1_mask, winsize, overlap, dt)
    mask_w = divide_windows(mask, windowsize=[winsize, winsize], step=winsize-overlap)[2] >= 0.5
    assert(mask_w.shape==x.shape)
    u[~mask_w] = np.nan
    v[~mask_w] = np.nan
    return x, y, u, v

def tangent_unit(point, center):
    """
    Compute tangent unit vector based on point coords and center coords.

    :param point: coordinates of the point of interest, 2-tuple
    :param center: coordinates of circle center, 2-tuple
    :return: tangent unit vector
    """
    point = np.array(point)
    # center = np.array(center)
    r = np.array((point[0] - center[0], point[1] - center[1]))
    # the following two lines set the initial value for the x of the tangent vector
    ind = np.logical_or(r[1] > 0, np.logical_and(r[1] == 0, r[0] > 0))
    x1 = np.ones(point.shape[1:])
    x1[ind] = -1
    y1 = np.zeros(point.shape[1:])
    x1[(r[1]==0)] = 0
    y1[(r[1]==0)&(r[0]>0)] = -1
    y1[(r[1]==0)&(r[0]<0)] = 1

    y1[r[1]!=0] = np.divide(x1 * r[0], r[1], where=r[1]!=0)[r[1]!=0]
    length = (x1**2 + y1**2) ** 0.5
    return np.divide(np.array([x1, y1]), length, out=np.zeros_like(np.array([x1, y1])), where=length!=0)

def apply_mask(pivData, mask):
    """
    Apply a mask on PIV data, by adding a boolean column "mask" to the original x, y, u, v data file. Valid velocities are labeled ``True`` while invalid velocities are labeled ``False``.

    :param pivData: PIV data (x, y, u, v)
    :type pivData: pandas.DataFrame
    :param mask: an image, preferrably binary, where large value denotes valid data and small value denote invalid data. The image will be converted to a boolean array by ``mask = mask > mask.mean()``.
    :type mask: 2D array
    :return: masked PIV data.
    :rtype: pandas.DataFrame

    .. rubric:: Test

    .. code-block:: python

       pivData = pd.read_csv("test_files/piv-test.csv")
       mask = io.imread("test_files/mask.tif")
       mpiv = apply_mask(pivData, mask)

       fig, ax = plt.subplots(nrows=1, ncols=2)
       ax[0].imshow(mask, cmap="gray")
       ax[0].quiver(pivData.x, pivData.y, pivData.u, pivData.v, color="red")
       ax[1].imshow(mask, cmap="gray")
       ax[1].quiver(mpiv.x, mpiv.y, mpiv.u, mpiv.v, color="red")

    .. rubric:: Edit

    * Nov 17, 2022 -- Initial commit.
    * Nov 30, 2022 -- Instead of replacing invalid data with ``np.nan``, add an additional column, where the validity of data is specified.
    * Dec 01, 2022 -- Remove the erosion step, since it is very obsecure to include this step here. If we want the mask to be more conservative (include less region to be sure that we are free from boundary effect), we can modify the mask in ImageJ and apply again on the PIV data.
    * Dec 19, 2022 -- Modify docstring to be consistent with code action.
    """
    mask = mask > mask.mean()
    ind = mask[pivData.y.astype("int"), pivData.x.astype("int")]
    pivData["mask"] = ind
    return pivData

# %% codecell
class piv_data:
    """Tools for PIV data downstream analysis, such as correlation, mean velocity,
    derivative fields, energy, enstrophy, energy spectrum, etc."""
    def __init__(self, file_list, fps=50, cutoff=250):
        """file_list: return value of readdata"""
        self.piv_sequence = file_list
        self.dt = 2 / fps # time between two data files
        self.stack = self.load_stack(cutoff=cutoff)
    def load_stack(self, cutoff=None):
        """
        Load PIV data in 3D numpy.array.

        :return: U, V -- PIV data stacked in 3D, with the first axis as time.
        """
        u_list = []
        v_list = []
        for num, i in self.piv_sequence.iterrows():
            x, y, u, v = read_piv(i.Dir)
            if num == 0:
                shape = x.shape
            else:
                if x.shape != shape:
                    break
            u_list.append(u)
            v_list.append(v)
            if cutoff is not None:
                if num > cutoff:
                    break
        return np.stack(u_list, axis=0), np.stack(v_list, axis=0)
    def vacf(self, mode="direct", smooth_method="gaussian", smooth_window=3, xlim=None, plot=False):
        """Compute averaged vacf from PIV data.
        This is a wrapper of function autocorr1d(), adding the averaging over all the velocity spots.
        Args:
        mode -- the averaging method, can be "direct" or "weighted".
                "weighted" will use mean velocity as the averaging weight, whereas "direct" uses 1.
        smooth_window -- window size for gaussian smoothing in time
        xlim -- xlim for plotting the VACF, does not affect the return value
        Returns:
        corrData -- DataFrame of (t, corrx, corry)
        Edit:
        Mar 23, 2022 -- add smoothn smoothing option
        Nov 15, 2022 -- Fix inconsistency with :py:func:`corrLib.autocorr1d`
        """
        # rearrange vstack from (f, h, w) to (f, h*w), then transpose
        corr_components = []
        for name, stack in zip(["corrx", "corry"], self.stack):
            stack_r = stack.reshape((stack.shape[0], -1)).T
            stack_r = stack_r[~np.isnan(stack_r).any(axis=1)]
            if smooth_method == "gaussian":
                stack_r = scipy.ndimage.gaussian_filter(stack_r, (0, smooth_window/4))
            elif smooth_method == "smoothn":
                stack_r = smoothn(stack_r, axis=1)[0]
            # compute autocorrelation
            corr_list = []
            weight = 1
            normalizer = 0
            for x in stack_r:
                if np.isnan(x[0]) == False: # masked out part has velocity as nan, which cannot be used for correlation computation
                    if mode == "weighted":
                        weight = abs(x).mean()
                    corr, lagt = autocorr1d(x, np.arange(len(x))*self.dt) * weight
                    if np.isnan(corr.sum()) == False:
                        normalizer += weight
                        corr_list.append(corr)
            corr_mean = np.nansum(np.stack(corr_list, axis=0), axis=0) / normalizer
            corr_components.append(pd.DataFrame({"c": corr_mean, "t": lagt}).set_index("t").rename(columns={"c": name}))
        ac = pd.concat(corr_components, axis=1)
        # plot autocorrelation functions
        if plot == True:
            fig, ax = plt.subplots(figsize=(3.5, 3), dpi=100)
            ax.plot(ac.index, ac.corrx, label="$C_x$")
            ax.plot(ac.index, ac.corry, label="$C_y$")
            ax.plot(ac.index, ac.mean(axis=1), label="mean", color="black")
            ax.set_xlabel("time (s)")
            ax.set_ylabel("VACF")
            ax.legend(frameon=False)
            ax.set_xlim([0, ac.index.max()])
            if xlim is not None:
                ax.set_xlim(xlim)
        return ac
    def corrS2d(self, mode="sample", n=10, plot=False):
        """Spatial correlation of velocity field.
        mode -- "sample" or "full"
                "sample" will sample n frames to compute the correlation
                "full" will sample all available frames to compute the correlation (could be computationally expensive)
        n -- number of frames to sample"""
        interval = max(len(self.piv_sequence) // n, 1)
        CV_list = []
        for num, i in self.piv_sequence[::interval].iterrows():
            x, y, u, v = read_piv(i.Dir)
            if num == 0:
                shape = x.shape
            else:
                if x.shape != shape:
                    break
            X, Y, CA, CV = corrS(x, y, u, v)
            CV_list.append(CV)
        CV_mean = np.stack(CV_list, axis=0).mean(axis=0)
        if plot == True:
            plt.imshow(CV_mean)
            plt.colorbar()
        return X, Y, CV_mean
    def corrS1d(self, mode="sample", n=10, xlim=None, plot=False):
        """Compute 2d correlation and convert to 1d. 1d correlation will be represented as pd.DataFrame of (R, C)."""
        X, Y, CV = self.corrS2d(mode=mode, n=n)
        dc = distance_corr(X, Y, CV)
        if plot == True:
            fig, ax = plt.subplots(figsize=(3.5, 3), dpi=100)
            ax.plot(dc.R, dc.C)
            ax.set_xlim([0, dc.R.max()])
            ax.set_xlabel("$r$ (pixel)")
            ax.set_ylabel("spatial correlation")
            if xlim is not None:
                ax.set_xlim(xlim)
        return dc
    def mean_velocity(self, mode="abs", plot=False):
        """Mean velocity time series.
        mode -- "abs" or "square"."""
        vm_list = []
        for num, i in self.piv_sequence.iterrows():
            x, y, u, v = read_piv(i.Dir)
            if mode == "abs":
                vm = np.nanmean((u ** 2 + v ** 2) ** 0.5)
            elif mode == "square":
                vm = np.nanmean((u ** 2 + v ** 2))  ** 0.5
            vm_list.append(vm)
        if plot == True:
            fig, ax = plt.subplots(figsize=(3.5, 3), dpi=100)
            ax.plot(np.arange(len(vm_list))*self.dt, vm_list)
            ax.set_xlabel("time (s)")
            ax.set_ylabel("mean velocity (px/s)")
        return pd.DataFrame({"t": np.arange(len(vm_list))*self.dt, "v_mean": vm_list})
    def order_parameter(self, center, mode="wioland"):
        """
        Compute order parameter of a velocity field. We currently have two definitions of order parameters, from Wioland 2013 and Hamby 2018, respectively. This function implements the calculations of both definitions.
        """
        def wioland2013(pivData, center):
            """Compute order parameter with PIV data and droplet center coords using the method from wioland2013.
            Args:
            pivData -- DataFrame of x, y, u, v
            center -- 2-tuple droplet center coords
            Return:
            OP -- float, max to 1
            """
            pivData = pivData.dropna()
            point = (pivData.x, pivData.y)
            tu = tangent_unit(point, center)
            # \Sigma vt
            sum_vt = abs((pivData.u * tu[0] + pivData.v * tu[1])).sum()
            sum_v = ((pivData.u**2 + pivData.v**2) ** 0.5).sum()
            OP = (sum_vt/sum_v - 2/np.pi) / (1 - 2/np.pi)
            return OP
        def hamby2018(pivData, center):
            """Computes order parameter using the definition in Hamby 2018.
            Args:
            pivData - DataFrame of x, y, u, v
            center - 2-tuple center of droplet
            Returns:
            OP - order parameter
            """
            pivData = pivData.dropna()
            tu = tangent_unit((pivData.x, pivData.y), center)
            pivData = pivData.assign(tu=tu[0], tv=tu[1])
            OP = (pivData.u * pivData.tu + pivData.v * pivData.tv).sum() / ((pivData.u ** 2 + pivData.v ** 2) ** 0.5).sum()
            return OP
        OP_list = []
        if mode == "wioland":
            for num, i in self.piv_sequence.iterrows():
                pivData = pd.read_csv(i.Dir)
                OP = wioland2013(pivData, center)
                OP_list.append(OP)
        elif mode == "hamby":
            for num, i in self.piv_sequence.iterrows():
                pivData = pd.read_csv(i.Dir)
                OP = hamby2018(pivData, center)
                OP_list.append(OP)
        return pd.DataFrame({"t": np.arange(len(OP_list)) * self.dt, "OP": OP_list})

class compact_PIV:
    """
    Compact PIV data structure. Instead of saving PIV data of each frame pair in separated text files, we can save them in a more compact form, where (x, y, mask) information are only saved once and only velocity informations are kept in 3D arrays. The data will be saved in a Matlab style .mat file, and the internal structure is a Python dictionary, with entries ("x", "y", "labels", "u", "v", "mask"). Since here x, y, u, v are no longer the same shape, accessing PIV data from a specific frame becomes less straight forward. This class is written to enable straightforward data access and saving. For a more detailed guide of using this class, see `compact_PIV tutorial <https://zloverty.github.io/code/tutorials/compact_PIV.html>`_.
    """
    def __init__(self, data):
        """
        Initialize compact_PIV object from data.

        :param data: can be dict or pandas.DataFrame. If it's dict, set the value directly to self.data. If it's DataFrame, construct a dict from the DataFrame (filelist).
        """
        if isinstance(data, dict):
            self.data = data
        elif isinstance(data, pd.DataFrame):
            if "Name" in data and "Dir" in data:
                self.data = self._from_filelist(data)
            else:
                raise ValueError
    def get_frame(self, i, by="index"):
        if by == "index":
            ind = i
        elif by == "filename":
            ind = self.get_labels().index(i)
        u, v = self.data["u"][ind], self.data["v"][ind]
        if "mask" in self.data.keys():
            u[~self.data["mask"].astype("bool")] = np.nan
            v[~self.data["mask"].astype("bool")] = np.nan
        return self.data["x"], self.data["y"], u, v
    def __repr__(self):
        return str(self.data)
    def __getitem__(self, indices):
        return self.data[indices]
    def _from_filelist(self, filelist):
        """
        Construct dict data from filelist of conventional PIV data.

        :param filelist: return value of readdata.
        """
        compact_piv = {}
        pivData = pd.read_csv(filelist.at[0, "Dir"])
        x, y, u, v = to_matrix(pivData)
        # set x, y values
        compact_piv["x"] = x
        compact_piv["y"] = y
        # set mask value, if exists
        if "mask" in pivData:
            mask_bool = np.reshape(np.array(pivData["mask"]), x.shape)
            compact_piv["mask"] = mask_bool
        # set u, v values and label value
        ul, vl = [], []
        label = []
        for num, i in filelist.iterrows():
            label.append(i.Name)
            pivData = pd.read_csv(i.Dir)
            x, y, u, v = to_matrix(pivData)
            ul.append(u)
            vl.append(v)
        compact_piv["u"] = np.stack(ul)
        compact_piv["v"] = np.stack(vl)
        compact_piv["labels"] = label
        return compact_piv
    def get_labels(self): # filenames originally used for constructing this data
        return list(self.data["labels"])
    def to_mat(self, fname):
        savemat(fname, self.data)
# %% codecell
if __name__ == '__main__':
    # %% codecell
    folder = r"test_images\moving_mask_piv\piv_result"
    l = readdata(folder, "csv")
    piv = piv_data(l, fps=50)
    # %% codecell
    vacf = piv.vacf(smooth_window=2, xlim=[0, 0.1])
    # autocorr1d(np.array([1,1,1]))
    # %% codecell
    corr1d = piv.corrS1d(n=600, xlim=[0, 170], plot=True)
    # %% codecell
    piv.mean_velocity(plot=True)
    # %% codecell
    op = piv.order_parameter((87, 87), mode="hamby")
    op



import numpy as np
from scipy import optimize
import matplotlib.pyplot as plt
import pandas as pd

def label_slope(x, y, location='n'):
    def power(x, a, b):
        return a*x**b    
    popt, pcov = optimize.curve_fit(power, x, y)
    # label the slope of a power law curve in a log-log plot with a dashed line
    # location indicates the relative location of the dashed line as to the original data line (only 'n' is available, other options need to be implemented)
    xmin = np.log10(x.min())
    xmax = np.log10(x.max())
    ymin = np.log10(y.min())
    ymax = np.log10(y.max())
    if location == 'n':
        xf = 10**np.array([xmin + (xmax - xmin)*3/8, xmin + (xmax - xmin)*5/8])
    elif location == 'nw':
        xf = 10**np.array([xmin + (xmax - xmin)*2/8, xmin + (xmax - xmin)*4/8])
    elif location == 'ne':
        xf = 10**np.array([xmin + (xmax - xmin)*4/8, xmin + (xmax - xmin)*6/8])
    yf = 10**((ymax-ymin)/7)*popt[0]*xf**popt[1]
    xfmin = np.log10(xf)[0]
    xfmax = np.log10(xf)[1]
    xt = 10**(xfmin + (xfmax - xfmin)*3/8)
    yt = 10**((ymax-ymin)/20)*10**np.log10(yf).mean()
    return xf, yf, xt, yt, popt[1]

def reformat_sequence(original):
    """
    This is not done yet. The basic working principles are implemented. Input and output are not well defined.
    """
    upg = 5 # unit per group
    gpr = 4 # group per row
    with open(r'I:\Github\Python\ForFun\DNA\test_files\flic_formatted.txt', 'w') as fw:
        pass
    with open(r'I:\Github\Python\ForFun\DNA\test_files\flic.txt', 'r') as f:
        seq = f.read().replace('\n', '')
        with open(r'I:\Github\Python\ForFun\DNA\test_files\flic_formatted.txt', 'a') as fw:
            count = 1
            line_count = 1
            for i in range(0, int(len(seq)/upg)+1):
                if count % gpr == 1:
                    fw.write('{:05d}  '.format((line_count-1)*upg*gpr))
                    line_count += 1
                fw.write(seq[upg*i:upg*i+upg])
                if count % gpr != 0:
                    fw.write(' ')
                else:
                    fw.write('\n')
                count += 1
    return reformatted

if __name__ == '__main__':
    data = pd.read_csv(r'I:\Github\Python\Correlation\test_images\GNF\stat\data.csv')
    data = data.loc[data.Name=='100-1']
    x = data.n
    y = data.d
    xf, yf, xt, yt, slope = label_slope(x, y, location='ne')
    plt.plot(x, y)
    plt.plot(xf, yf, ls='--', color='black')
    plt.text(xt, yt, '{:.2f}'.format(slope))
    plt.xscale('log')
    plt.yscale('log')
    plt.show()


# %% codecell
import os
import numpy as np
from scipy import fft
from scipy.signal import medfilt2d, convolve2d, fftconvolve
from scipy.optimize import curve_fit
import pandas as pd
import psutil
from skimage import io
import time
import shutil
from nd2reader import ND2Reader
# %% codecell

def dirrec(path, filename):
    """
    Recursively look for all the directories of files with name *filename*.

    :param path: the directory where you want to look for files.
    :type path: str
    :param filename: name of the files you want to look for.
    :type filename: str
    :return: a list of full directories of files with name *filename*
    :rtype: list[str]

    .. note::

       :code:`filename` can be partially specified, e.g. :code:`*.py` to search for all the files that end with *.py*. Similarly, setting :code:`filename` as :code:`*track*` will search for all files starting with *track*.

    .. rubric:: EDIT

    :11152022: Fix a bug, which falsely uses :py:func:`dirrec` within itself to iterate over subdirectories.
    """
    dirList = []
    for r, d, f in os.walk(path):
        # for dir in d:
        #     tmp = dirrec(dir, filename)
        #     if tmp:
        #         dirList.append(tmp)
        for file in f:
            if filename.startswith('*'):
                if file.endswith(filename[1:]):
                    dirList.append(os.path.join(r, file))
            elif filename.endswith('*'):
                if file.startswith(filename[:-1]):
                    dirList.append(os.path.join(r, file))
            elif file == filename:
                dirList.append(os.path.join(r, file))
    return dirList

def to8bit(img16):
    """
    Enhance contrast and convert to 8-bit.

    :param img: mono image of any dtype
    :type img: 2d array
    :return: 8-bit image
    :rtype: uint8 2d array
    """
    # if img16.dtype != 'uint16':
        # raise ValueError('16-bit grayscale image is expected')
    maxx = img16.max()
    minn = img16.min()
    img8 = (img16 - minn) / (maxx - minn) * 255
    return img8.astype('uint8')

def bpass(*args):
    """
    Apply bandpass filter on images. Useful when raw images have long wavelength intensity gradient.

    :param img: 8-bit image
    :type img: 2d array
    :param low: lower limit wavelength
    :type low: int
    :param high: upper limit wavelength
    :type high: int
    :return: processed image with low and high wavelength signals filtered
    :rtype: 2d array
    """
    img8 = args[0]
    low = args[1]
    high = args[2]
    def gen_filter(img, low, high):
        filt = np.zeros(img.shape)
        h, w = img.shape
        center = [int(w/2), int(h/2)]
        Y, X = np.ogrid[:h, :w]
        dist = ((X - center[0])**2 + (Y-center[1])**2)**.5

        filt[(dist>low)&(dist<=high)] = 1
        return filt
    filt = gen_filter(img8, low, high)
    filt = fft.ifftshift(filt)
    im_fft = fft.fft2(img8)
    im_fft_filt = im_fft * filt
    im_new = fft.ifft2(im_fft_filt).real
    im_new = im_new - im_new.min()
    im_new = np.floor_divide(im_new, (im_new.max()+1)/256)
    return im_new.astype('uint8')

def bestcolor(n):
    """
    Default plot color scheme of Matplotlib and Matlab. It is the same as `the "tab10" colormap of Matplotlib.colormaps <https://matplotlib.org/stable/tutorials/colors/colormaps.html#qualitative>`_.

    :param n: integer from 0 to 9, specifying the index of the color in the list
    :type n: int
    :return: the hex code of the specified color
    :rtype: str
    """
    colors = ['#1f77b4', '#ff7f0e', '#2ca02c', '#d62728', '#9467bd',
              '#8c564b', '#e377c2', '#7f7f7f', '#bcbd22', '#17becf']
    return colors[n]

def wowcolor(n):
    """
    WOW class color scheme, used in my density fluctuations paper. I used to think these colors are aesthetically pleasing, but now I feel they are too saturated and cause eye fatigue easily. Therefore I would avoid using these colors in future publications.

    :param n: integer from 0 to 9, specifying the index of the color in the list
    :type n: int
    :return: the hex code of the specified color
    :rtype: str
    """
    colors = ['#C41F3B', '#A330C9', '#FF7D0A', '#A9D271', '#40C7EB',
              '#00FF96', '#F58CBA', '#FFF569', '#0070DE', '#8787ED',
              '#C79C6E', '#BBBBBB', '#1f77b4', '#ff7f0e', '#2ca02c',
              '#d62728', '#9467bd', '#8c564b', '#e377c2', '#7f7f7f',
              '#bcbd22', '#17becf']
    return colors[n]

def matlab_style_gauss2D(shape=(3,3),sigma=0.5):
    """
    Generate a 2D gaussian mask - should give the same result as MATLAB's :code:`fspecial('gaussian',[shape],[sigma])`.

    :param shape: shape of the mask, default to (3,3)
    :type shape: tuple
    :param simga: standard deviation of the mask, default to 0.5
    :type sigma: float
    :return: a gaussian mask
    :rtype: 2d array
    """
    m,n = [(ss-1.)/2. for ss in shape]
    y,x = np.ogrid[-m:m+1,-n:n+1]
    h = np.exp( -(x*x + y*y) / (2.*sigma*sigma) )
    h[ h < np.finfo(h.dtype).eps*h.max() ] = 0
    sumh = h.sum()
    if sumh != 0:
        h /= sumh
    return h

def FastPeakFind(data):
    """
    Detect peak in 2D images.

    I rewrote a `Matlab function <https://github.com/uw-cmg/MATLAB-loop-detection/blob/master/FastPeakFind.m>`_ with the same name in Python. The function, in my opinion, is unnecessarily complex, with thresholding, filtering, edge excluding etc. in the same function, making it very long and not easy to read. Moreover, it sometimes fails obviously simple tasks. Therefore, I would use :code:`skimage.feature.peak_local_max` for the same purpose, whenever possible.

    :param data: 2d images to find peaks in
    :type data: 2d array
    :return: coordinates of peaks in an Nx2 array
    :rtype: 2d array
    """
    if str(data.dtype) != 'float32':
        data = data.astype('float32')
    mf = medfilt2d(data, kernel_size=3)
    mf = mf.astype('float32')
    thres = max(min(np.amax(mf,axis=0)), min(np.amax(mf,axis=1)))
    filt = matlab_style_gauss2D()
    conv = convolve2d(mf, filt, mode='same')
    w_idx = conv > thres
    bw = conv.copy()
    bw[w_idx] = 1
    bw[~w_idx] = 0
    thresholded = np.multiply(bw, conv)
    edg = 3
    shape = data.shape
    idx = np.nonzero(thresholded[edg-1: shape[0]-edg-1, edg-1: shape[1]-edg-1])
    idx = np.transpose(idx)
    cent = []
    for xy in idx:
        x = xy[0]
        y = xy[1]
        if thresholded[x, y] >= thresholded[x-1, y-1] and \
            thresholded[x, y] > thresholded[x-1, y] and \
            thresholded[x, y] >= thresholded[x-1, y+1] and \
            thresholded[x, y] > thresholded[x, y-1] and \
            thresholded[x, y] > thresholded[x, y+1] and \
            thresholded[x, y] >= thresholded[x+1, y-1] and \
            thresholded[x, y] > thresholded[x+1, y] and \
            thresholded[x, y] >= thresholded[x+1, y+1]:
            cent.append(xy)
    cent = np.asarray(cent).transpose()
    return cent

def minimal_peakfind(img):
    """
    Minimal version of :py:func:`myImageLib.FastPeakFind`. Remove all the preprocessings, such as median filter, gauss filter and thresholding. Only keep the essential peak finding functionality.

    :param img: input image
    :type img: 2-d array
    :return: coordinates of peaks in an Nx2 array
    :rtype: 2d array
    """
    edg = 3
    shape = img.shape
    idx = np.nonzero(img[edg-1: shape[0]-edg-1, edg-1: shape[1]-edg-1])
    idx = np.transpose(idx)
    cent = []
    for xy in idx:
        x = xy[0]
        y = xy[1]
        if img[x, y] >= img[x-1, y-1] and \
            img[x, y] > img[x-1, y] and \
            img[x, y] >= img[x-1, y+1] and \
            img[x, y] > img[x, y-1] and \
            img[x, y] > img[x, y+1] and \
            img[x, y] >= img[x+1, y-1] and \
            img[x, y] > img[x+1, y] and \
            img[x, y] >= img[x+1, y+1]:
            cent.append(xy)
    cent = np.asarray(cent).transpose()
    return cent

def maxk(array, num_max):
    """
    Sort a numerical array and return the maximum :code:`num_max` elements.

    :param array: input array
    :type array: 1-d numerical array
    :param num_max: number of maximal elements to return
    :type num_max: int
    :return: index of maximal elements
    :rtype: int array
    """
    array = np.asarray(array)
    length = array.size
    array = array.reshape((1, length))
    idx = np.argsort(array)
    idx2 = np.flip(idx)
    return idx2[0, 0: num_max]

def track_spheres_dt(img, num_particles):
    """
    Use correlation tracking method to find spheres in an image.
    """
    def gauss1(x,a,x0,sigma):
        return a*np.exp(-(x-x0)**2/(2*sigma**2))
    cent = FastPeakFind(img)
    num_particles = min(num_particles, cent.shape[1])
    peaks = img[cent[0], cent[1]]
    ind = maxk(peaks, num_particles)
    max_coor_tmp = cent[:, ind]
    max_coor = max_coor_tmp.astype('float32')
    pk_value = peaks[ind]
    for num in range(0, num_particles):
        try:
            x = max_coor_tmp[0, num]
            y = max_coor_tmp[1, num]
            fitx1 = np.asarray(range(x-7, x+8))
            fity1 = np.asarray(img[range(x-7, x+8), y])
            popt,pcov = curve_fit(gauss1, fitx1, fity1, p0=[1, x, 3])
            max_coor[0, num] = popt[1]
            fitx2 = np.asarray(range(y-7, y+8))
            fity2 = np.asarray(img[x, range(y-7, y+8)])
            popt,pcov = curve_fit(gauss1, fitx2, fity2, p0=[1, y, 3])
            max_coor[1, num] = popt[1]
        except:
            print('Fitting failed')
            max_coor[:, num] = max_coor_tmp[:, num]
            continue
    return max_coor, pk_value

def gauss1(x,a,x0,sigma,b):
    """
    1-d gaussian function. Usually used for fitting.
    """
    return a*np.exp(-(x-x0)**2/(2*sigma**2)) + b

def show_progress(progress, label='', bar_length=60):
    """
    Display a progress bar in command line environment, which looks like

    .. code-block:: console

      label [##############-----------] 62%

    This is a useful tool for tracking work progress in a batch processing task on a server.

    :param progress: the progress of the work. It is a number between 0 and 1, where 0 is start and 1 is finish
    :type progress: float
    :param label: a string to put before the progress bar. It can be the name of the work, or the actual number of files that have been processed, and so on. Default to :code:`''`.
    :type label: str
    :param bar_length: length of the progress bar, in the unit of characters. Default to :code:`60`.
    :type bar_length: int
    :return: None
    """
    N_finish = int(progress*bar_length)
    N_unfinish = bar_length - N_finish
    print('{0} [{1}{2}] {3:.1f}%'.format(label, '#'*N_finish, '-'*N_unfinish, progress*100), end="\r")

def readdata(folder, ext='csv', mode="i"):
    """
    Wrapper of :py:func:`dirrec`, but easier to use when reading one type of files in a given folder. Instead of returning a list of directories as :py:func:`dirrec` does, :py:func:`readdata` puts the file names and corresponding full directories in a :code:`pandas.DataFrame`. The table will be sorted by the file names (strings), so the order would likely be correct. In the worst case, it is still easier to resort the :code:`pandas.DataFrame`, compared to the list of strings returned by :py:func:`dirrec`.

    :param folder: the folder to read files from
    :type folder: str
    :param ext: optional param, default to "csv", specifies the extension of files to be read
    :type ext: str
    :param mode: "i" for immediate, "r" for recursive. Default to "i"
    :type mode: str
    :return: a 2-column table containing file names and the corresponding full directories
    :rtype: pandas.DataFrame

    .. rubric:: EDIT

    :11152022: Add mode optional argument, to specify whether to read data only in the immediate folder, or read recursively.
    """
    dataDirs = dirrec(folder, '*.' + ext)
    dataDirsCopy = dataDirs.copy()
    if mode == "i":
        for dataDir in dataDirsCopy:
            relpath = dataDir.replace(folder, "").strip(os.sep)
            if os.sep in relpath:
                dataDirs.remove(dataDir)
    nameList = []
    dirList = []
    for dataDir in dataDirs:
        path, file = os.path.split(dataDir)
        name, ext = os.path.splitext(file)
        nameList.append(name)
        dirList.append(dataDir)
    fileList = pd.DataFrame()
    fileList = fileList.assign(Name=nameList, Dir=dirList)
    fileList = fileList.sort_values(by=['Name']).reset_index(drop=True)
    return fileList

def normxcorr2(template, image, mode="full"):
    """
    Compute normalized cross-correlation map between an image and a template. Input arrays should be floating point numbers.

    :param template: N-D array, of template or filter you are using for cross-correlation. Must be less or equal dimensions to image. Length of each dimension must be less than length of image.
    :type template: float array
    :param image: N-D array
    :type image: float array
    :param mode: Options, "full", "valid", "same". full (Default): The output of fftconvolve is the full discrete linear convolution of the inputs. Output size will be image size + 1/2 template size in each dimension. valid: The output consists only of those elements that do not rely on the zero-padding. same: The output is the same size as image, centered with respect to the ‘full’ output.
    :return: N-D array of same dimensions as image. Size depends on mode parameter.
    :rtype: float array
    """

    # If this happens, it is probably a mistake
    if np.ndim(template) > np.ndim(image) or \
            len([i for i in range(np.ndim(template)) if template.shape[i] > image.shape[i]]) > 0:
        print("normxcorr2: TEMPLATE larger than IMG. Arguments may be swapped.")

    template = template - np.mean(template)
    image = image - np.mean(image)

    a1 = np.ones(template.shape)
    # Faster to flip up down and left right then use fftconvolve instead of scipy's correlate
    ar = np.flipud(np.fliplr(template))
    out = fftconvolve(image, ar.conj(), mode=mode)

    image = fftconvolve(np.square(image), a1, mode=mode) - \
            np.square(fftconvolve(image, a1, mode=mode)) / (np.prod(template.shape))

    # Remove small machine precision errors after subtraction
    image[np.where(image < 0)] = 0

    template = np.sum(np.square(template))
    out = out / np.sqrt(image * template)

    # Remove any divisions by 0 or very close to 0
    out[np.where(np.logical_not(np.isfinite(out)))] = 0

    return out

def match_hist(im1, im2):
    """
    Match the histogram of im1 to that of im2

    :param im1: image
    :type im1: 2d array
    :param im2: image
    :type im2: 2d array
    :return: a modified version of im1, that matches im2's histogram
    :rtype: 2d array

    .. rubric:: EDIT

    :11142022: Move from ``corrLib`` to ``myImageLib``.
    """
    return (abs(((im1 - im1.mean()) / im1.std() * im2.std() + im2.mean()))+1).astype('uint8')

def xy_bin(xo, yo, n=100, mode='log', bins=None):
    """
    Bin x, y data on log or linear scale


    :param xo: input x
    :param yo: input y
    :param n: points after binning
    :param mode: "lin" or "log", scale to bin the data
    :param bins: set the bins to bin data together

    :return:
        * x: binned x
        * y: means in bins

    .. rubric: Edit

    :11042020: Change function name to xy_bin, to incorporate the mode parameter, so that the function can do both log space binning and linear space binning.
    :11172020: add bins kwarg, allow user to enter custom bins.
    :12162021: fix divided by 0 issue.
    :11142022: Move from ``corrLib`` to ``myImageLib``.
    """
    assert(len(xo)==len(yo))
    if bins is None:
        if mode == 'log':
            x = np.logspace(np.log10(xo[xo>0].min()), np.log10(xo.max()), n+1)
        elif mode == 'lin':
            x = np.linspace(xo.min(), xo.max(), n+1)
    else:
        x = np.sort(bins)
    top = np.histogram(xo, x, weights=yo)[0]
    bot = np.histogram(xo, x)[0]
    ind = bot > 0
    xb = ((x[1:] + x[:-1]) / 2)[ind]
    yb = top[ind] / bot[ind]
    return xb, yb
# %% codecell

class rawImage:
    """
    This class converts raw images to tif sequences. Throughout my research, I have mostly worked with two raw image formats: *\*.nd2* and *\*.raw*. Typically, they are tens of GB large, and are not feasible to load into the memory of a PC as a whole. Therefore, the starting point of my workflow is to convert raw images into sequences of *\*.tif* images. """

    def __init__(self, file_dir):
        """
        Construct rawImage object using the file directory.
        """
        self.file = file_dir
        self.type = os.path.splitext(self.file)[1]
        if self.type == ".nd2":
            with ND2Reader(self.file) as images:
                self.images = images
        elif self.type == ".raw":
            print("raw image handling is pending implementation.")
        else:
            raise ValueError
    def __repr__(self):
        """Ideally, I can see some general informations about this image. For example, the number of frames and the frame size. Frame rate would also be helpful."""
        repr_str = "source file: {0}\nimage shape: {1}".format(self.file, self.images.shape)
        return repr_str
    def extract_tif(self):
        """Wrapper of all format-specific extractors."""
        file, ext = os.path.splitext(self.file)
        if ext == ".raw":
            self._extract_raw()
        elif ext == ".nd2":
            self._extract_nd2()
        else:
            raise ValueError("Unrecognized image format {}".format(ext))
    def _extract_raw(self, cutoff=None):
        """Extract tif sequence from *\*.raw* file.
        :param cutoff: number of images to extract

        .. rubric:: Edit

        * Dec 07, 2022: fix progress bar error.
        """
        # read info from info_file
        folder, file = os.path.split(self.file)
        info_file = os.path.join(folder, "RawImageInfo.txt")
        if os.path.exists(info_file):
            fps, h, w = self.read_raw_image_info(info_file)
        else:
            print("Info file missing!")

        # create output folders, skip if both folders exist
        save_folder = folder
        out_raw_folder = os.path.join(save_folder, 'raw')
        out_8_folder = os.path.join(save_folder, '8-bit')
        if os.path.exists(out_raw_folder) and os.path.exists(out_8_folder):
            print(time.asctime() + " // {} tif folders exists, skipping".format(self.file))
            return None
        if os.path.exists(out_raw_folder) == False:
            os.makedirs(out_raw_folder)
        if os.path.exists(out_8_folder) == False:
            os.makedirs(out_8_folder)

        # check disk
        if self._disk_capacity_check() == False:
            raise SystemError("No enough disk capacity!")

        # calculate buffer size based on available memory, use half of it
        file_size = os.path.getsize(self.file)
        avail_mem = psutil.virtual_memory().available
        unit_size = (h * w + 2) * 2 # the size of a single unit in bytes (frame# + image)
        buffer_size = ((avail_mem // 2) // unit_size) * unit_size # integer number of units
        remaining_size = file_size

        # print("Memory information:")
        # print("Available memory: {:.1f} G".format(avail_mem / 2 ** 30))
        # print("File size: {:.1f} G".format(file_size / 2 ** 30))
        # print("Buffer size: {:.1f} G".format(buffer_size / 2 ** 30))

        # read the binary files, in partitions if needed
        num = 0
        n_images = int(file_size // unit_size)
        t0 = time.monotonic()
        while remaining_size > 0:
            # load np.array from buffer
            if remaining_size > buffer_size:
                read_size = buffer_size
            else:
                read_size = remaining_size
            with open(self.file, "rb") as f:
                f.seek(-remaining_size, 2) # offset bin file reading, counting remaining size from EOF
                bytestring = f.read(read_size)
                a = np.frombuffer(bytestring, dtype=np.uint16)
            remaining_size -= read_size

            # manipulate the np.array
            assert(a.shape[0] % (h*w+2) == 0)
            num_images = a.shape[0] // (h*w+2)
            img_in_row = a.reshape(num_images, h*w+2)
            labels = img_in_row[:, :2] # not in use currently
            images = img_in_row[:, 2:]
            images_reshape = images.reshape(num_images, h, w)

            # save the image sequence
            for label, img in zip(labels, images_reshape):
                # num = label[0] + label[1] * 2 ** 16 + 1 # convert image label to uint32 to match the info in StagePosition.txt
                io.imsave(os.path.join(save_folder, 'raw', '{:05d}.tif'.format(num)), img, check_contrast=False)
                io.imsave(os.path.join(save_folder, '8-bit', '{:05d}.tif'.format(num)), to8bit(img), check_contrast=False)
                t1 = time.monotonic() - t0
                show_progress((num+1 / n_images), label="{:.1f} frame/s".format(num / t1))
                num += 1
                if cutoff is not None:
                    if num > cutoff:
                        return None

    def read_raw_image_info(self, info_file):
        """
        Read image info, such as fps and image dimensions, from *\*.RawImageInfo.txt*.
        Helper function of :py:func:`myImageLib.rawImage.extract_raw`.
        """
        with open(info_file, 'r') as f:
            a = f.read()
        fps, h, w = a.split('\n')[0:3]
        return int(fps), int(h), int(w)

    def _extract_nd2(self):
        """
        Extract tif sequence from *\*.nd2* file.
        """
        # check disk
        if self._disk_capacity_check() == False:
            raise SystemError("No enough disk capacity!")

        folder, file = os.path.split(self.file)
        name, ext = os.path.splitext(file)
        saveDir = os.path.join(folder, name, 'raw')
        saveDir8 = os.path.join(folder, name, '8-bit')
        if os.path.exists(saveDir) == False:
            os.makedirs(saveDir)
        if os.path.exists(saveDir8) == False:
            os.makedirs(saveDir8)
        t0 = time.monotonic()
        with ND2Reader(self.file) as images:
            n_images = len(images)
            for num, image in enumerate(images):
                io.imsave(os.path.join(saveDir8, '{:05d}.tif'.format(num)), to8bit(image))
                io.imsave(os.path.join(saveDir, '%05d.tif' % num), image, check_contrast=False)
                t1 = time.monotonic() - t0
                show_progress((num+1) / n_images, label="{:.1f} frame/s".format(num/t1))

    def _disk_capacity_check(self):
        """Check if the capacity of disk is larger than twice of the file size.
        Args:
        file -- directory of the (.nd2) file being converted
        Returns:
        flag -- bool, True if capacity is enough."""
        d = os.path.split(self.file)[0]
        fs = os.path.getsize(self.file) / 2**30
        ds = shutil.disk_usage(d)[2] / 2**30
        print("File size {0:.1f} GB, Disk size {1:.1f} GB".format(fs, ds))
        return ds > 2 * fs